from typing import Callable, List, Optional, Tuple, TypeVar, Union, cast, overload

from mygrad.tensor_base import Tensor, implements_numpy_override
from mygrad.typing import ArrayLike, Shape

from .ops import *

__all__ = [
    "reshape",
    "squeeze",
    "ravel",
    "expand_dims",
    "broadcast_to",
    "atleast_1d",
    "atleast_2d",
    "atleast_3d",
]

_T = TypeVar("_T")


@implements_numpy_override()
def reshape(
    a: ArrayLike, newshape: Union[int, Shape], *, constant: Optional[bool] = None
) -> Tensor:
    """Returns a tensor with a new shape, without changing its data.

    This docstring was adapted from ``numpy.reshape``

    Parameters
    ----------
    a : ArrayLike
        The tensor to be reshaped

    newshape : Union[int, Tuple[int, ...]]
        The new shape should be compatible with the original shape. If
        an integer, then the result will be a 1-D tensor of that length.
        One shape dimension can be -1. In this case, the value is
        inferred from the length of the tensor and remaining dimensions.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor
        ``a`` with its shape changed permuted.  A new tensor is returned.

    Notes
    -----
    ``reshape`` utilizes C-ordering, meaning that it reads & writes elements using
    C-like index ordering; the last axis index changing fastest, and, proceeding
    in reverse order, the first axis index changing slowest.

    Examples
    --------
    >>> import mygrad as mg
    >>> a = mg.Tensor([[1,2,3], [4,5,6]])
    >>> mg.reshape(a, 6)
    Tensor([1, 2, 3, 4, 5, 6])

    >>> mg.reshape(a, (3,-1))   # the unspecified value is inferred to be 2
    Tensor([[1, 2],
            [3, 4],
            [5, 6]])"""
    return Tensor._op(Reshape, a, op_args=(newshape,), constant=constant)


@implements_numpy_override()
def squeeze(
    a: ArrayLike,
    axis: Optional[Union[int, Tuple[int, ...]]] = None,
    *,
    constant: Optional[bool] = None
) -> Tensor:
    """
    Remove single-dimensional entries from the shape of a tensor.

    This docstring was adapted from ``numpy.squeeze``

    Parameters
    ----------
    a : ArrayLike
        The tensor to be reshaped

    axis : Optional[int, Tuple[int, ...]]
        Selects a subset of the single-dimensional entries in the
        shape. If an axis is selected with shape entry greater than
        one, an error is raised.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor

    Raises
    ------
    ValueError
        If ``axis`` is not ``None``, and an axis being squeezed is not of length 1

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.Tensor([[[0], [1], [2]]])
    >>> x.shape
    (1, 3, 1)
    >>> mg.squeeze(x).shape
    (3,)
    >>> mg.squeeze(x, axis=0).shape
    (3, 1)
    >>> mg.squeeze(x, axis=1).shape
    Traceback (most recent call last) -> Tensor:
    ...
    ValueError: cannot select an axis to squeeze out which has size not equal to one
    >>> mg.squeeze(x, axis=2).shape
    (1, 3)"""
    return Tensor._op(Squeeze, a, op_args=(axis,), constant=constant)


@implements_numpy_override()
def ravel(a: ArrayLike, *, constant: Optional[bool] = None) -> Tensor:
    """
    Flattens contents of a tensor into a contiguous 1-D array.  A copy is made only if needed.

    This docstring was adapted from ``numpy.ravel``.

    Parameters
    ----------
    a : ArrayLike
        The tensor to be flattened

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor

    Notes
    -----
    ``ravel`` utilizes C-ordering, meaning that it reads & writes elements using
    C-like index ordering; the last axis index changing fastest, and, proceeding
    in reverse order, the first axis index changing slowest.

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.Tensor([[1, 2],
    ...                [3, 4]])
    >>> mg.ravel(x)
    Tensor([1, 2, 3, 4])
    """
    return Tensor._op(Ravel, a, constant=constant)


@implements_numpy_override()
def expand_dims(a: ArrayLike, axis: int, *, constant: Optional[bool] = None) -> Tensor:
    """
    Expand the dimensions of a tensor by adding a new axis.

    This docstring was adapted from ``numpy.expand_dims``.

    Parameters
    ----------
    a : ArrayLike
        The tensor to be expanded

    axis : int
        The position of the new axis in the expanded array shape.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.Tensor([1, 2])
    >>> x.shape
    (2,)
    >>> y = mg.expand_dims(x, 1)
    >>> y.shape
    (2, 1)
    >>> z = mg.expand_dims(y, 0)
    >>> z.shape
    (1, 2, 1)
    """
    return Tensor._op(ExpandDims, a, op_args=(axis,), constant=constant)


@implements_numpy_override()
def broadcast_to(
    a: ArrayLike, shape: Shape, *, constant: Optional[bool] = None
) -> Tensor:
    """
    Broadcast a tensor to a new shape.

    This docstring was adapted from ``numpy.broadcast_to``.

    Parameters
    ----------
    a : ArrayLike
        The tensor to be broadcasted

    shape: Tuple[int, ...]
        The shape of the broadcasted tensor. This shape
        should be broadcast-compatible with the original
        shape.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor

    Raises
    ------
    ValueError
        If the array is not compatible with the new shape
        according to Numpy's broadcasting rules.

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.Tensor([1, 2, 3])
    >>> mg.broadcast_to(x, (3,3))
    Tensor([[1, 2, 3],
            [1, 2, 3],
            [1, 2, 3]])
    >>> mg.broadcast_to(x, (4,4))
    Traceback (most recent call last) -> Tensor:
    ...
    ValueError: operands could not be broadcast together with remapped
    shapes [original->remapped]: (3,) and requested shape (4,4)
    """
    return Tensor._op(BroadcastTo, a, op_args=(shape,), constant=constant)


def _dispatch_atleast_kd(func: Callable[..., _T], Op, *tensors, k: int, constant) -> _T:
    if len(tensors) == 1:
        (t,) = tensors
        if (
            isinstance(t, Tensor)
            and t.ndim >= k
            and (constant is None or t.constant is constant)
        ):
            # return tensor unchanged
            return cast(_T, t)
        return cast(_T, Tensor._op(Op, t, constant=constant))
    else:
        out = [func(t, constant=constant) for t in tensors]
        return cast(_T, out)


@overload
def atleast_1d(
    tensors: ArrayLike, *, constant: Optional[bool] = None
) -> Tensor:  # pragma: no cover
    ...


@overload
def atleast_1d(
    *tensors: ArrayLike, constant: Optional[bool] = None
) -> List[Tensor]:  # pragma: no cover
    ...


@implements_numpy_override()
def atleast_1d(
    *tensors: ArrayLike, constant: Optional[bool] = None
) -> Union[Tensor, List[Tensor]]:
    """
    Convert inputs to tensors with at least one dimension.

    Scalar inputs are converted to 1-dimensional tensors, whilst
    higher-dimensional inputs are preserved.

    This docstring was adapted from ``numpy.atleast_1d``.

    Parameters
    ----------
    tens1, tens2, ... : ArrayLike
        One or more input tensors.

    Returns
    -------
    ret : Tensor | List[Tensor]
        A tensor, or list of tensors, each with ``a.ndim >= 1``.
        Copies are made only if necessary.

    See Also
    --------
    atleast_2d, atleast_3d

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.atleast_1d(1.0)
    array([1.])

    >>> x = mg.arange(9.0).reshape(3,3)
    >>> np.atleast_1d(x)
    Tensor([[0., 1., 2.],
            [3., 4., 5.],
            [6., 7., 8.]])
    >>> mg.atleast_1d(x) is x
    True

    >>> mg.atleast_1d(1, [3, 4])
    [Tensor([1]), Tensor([3, 4])]

    ``numpy.atleast_1d`` will dispatch appropriately on tensors.

    >>> x = mg.tensor(2.)
    >>> np.atleast_1d(x)
    Tensor([2.])

    >>> np.atleast_1d(x).backward()
    >>> x.grad
    array(1.)

    If any argument to ``numpy.atleast_1d`` is a Tensor, ``mygrad.atleast_1d``
    will be dispatched on all of the arguments.

    >>> np.atleast_1d(x, 1.)
    [Tensor([2.]), Tensor([1.])]
    """
    return _dispatch_atleast_kd(atleast_1d, AtLeast1D, *tensors, k=1, constant=constant)


@overload
def atleast_2d(
    tensors: ArrayLike, *, constant: Optional[bool] = None
) -> Tensor:  # pragma: no cover
    ...


@overload
def atleast_2d(
    *tensors: ArrayLike, constant: Optional[bool] = None
) -> List[Tensor]:  # pragma: no cover
    ...


@implements_numpy_override()
def atleast_2d(
    *tensors: ArrayLike, constant: Optional[bool] = None
) -> Union[Tensor, List[Tensor]]:
    """
    Convert inputs to tensors with at least one dimension.

    Scalar inputs are converted to 2-dimensional tensors, whilst
    higher-dimensional inputs are preserved.

    This docstring was adapted from ``numpy.atleast_2d``.

    Parameters
    ----------
    tens1, tens2, ... : ArrayLike
        One or more input tensors.

    Returns
    -------
    ret : Tensor | List[Tensor]
        A tensor, or list of tensors, each with ``a.ndim >= 2``.
        Copies are made only if necessary.

    See Also
    --------
    atleast_1d, atleast_3d

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.atleast_2d(3.0)
    Tensor([[3.]])

    >>> x = mg.arange(3.0)
    >>> mg.atleast_2d(x)
    array([[0., 1., 2.]])
    >>> mg.atleast_2d(x).base is x
    True

    >>> mg.atleast_2d(1, [1, 2], [[1, 2]])
    [Tensor([[1]]), Tensor([[1, 2]]), Tensor([[1, 2]])]

    ``numpy.atleast_2d`` will dispatch appropriately on tensors.

    >>> x = mg.tensor(2.)
    >>> np.atleast_2d(x)
    Tensor([[2.]])

    >>> np.atleast_2d(x).backward()
    >>> x.grad
    array(1.)

    If any argument to ``numpy.atleast_2d`` is a Tensor, ``mygrad.atleast_2d``
    will be dispatched on all of the arguments.

    >>> np.atleast_2d(x, 1.)
    [Tensor([[2.]]), Tensor([[1.]])]
    """
    return _dispatch_atleast_kd(atleast_2d, AtLeast2D, *tensors, k=2, constant=constant)


@overload
def atleast_3d(
    tensors: ArrayLike, *, constant: Optional[bool] = None
) -> Tensor:  # pragma: no cover
    ...


@overload
def atleast_3d(
    *tensors: ArrayLike, constant: Optional[bool] = None
) -> List[Tensor]:  # pragma: no cover
    ...


@implements_numpy_override()
def atleast_3d(
    *tensors: ArrayLike, constant: Optional[bool] = None
) -> Union[Tensor, List[Tensor]]:
    """
    Convert inputs to tensors with at least one dimension.

    Scalar inputs are converted to 3-dimensional tensors, whilst
    higher-dimensional inputs are preserved.

    This docstring was adapted from ``numpy.atleast_3d``.

    Parameters
    ----------
    tens1, tens2, ... : ArrayLike
        One or more input tensors.

    Returns
    -------
    ret : Tensor | List[Tensor]
        A tensor, or list of tensors, each with ``a.ndim >= 3``.
        Copies are made only if necessary. For example, a 1-D tensor of shape ``(N,)``
        becomes a view of shape ``(1, N, 1)``, and a 2-D tensor of shape ``(M, N)``
        becomes a view of shape ``(M, N, 1)``.

    See Also
    --------
    atleast_1d, atleast_3d

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.atleast_3d(3.0)
    Tensor([[[3.]]])

    >>> x = mg.arange(3.0)
    >>> mg.atleast_3d(x).shape
    (1, 3, 1)
    >>> mg.atleast_3d(x).base is x
    True

    >>> x = mg.arange(12.0).reshape(4,3)
    >>> mg.atleast_3d(x).shape
    (4, 3, 1)

    >>> mg.atleast_3d(1, [[1, 2]], [[[[1, 2]]]])
    [Tensor([[[1]]]), Tensor([[[1, 2]]]), Tensor([[[[1, 2]]]])]

    ``numpy.atleast_3d`` will dispatch appropriately on tensors.

    >>> x = mg.tensor(2.)
    >>> np.atleast_3d(x)
    Tensor([[[2.]]])

    >>> np.atleast_3d(x).backward()
    >>> x.grad
    array(1.)

    If any argument to ``numpy.atleast_3d`` is a Tensor, ``mygrad.atleast_3d``
    will be dispatched on all of the arguments.

    >>> np.atleast_3d(x, 1.)
    [Tensor([[[2.]]]), Tensor([[[1.]]])]
    """
    return _dispatch_atleast_kd(atleast_3d, AtLeast3D, *tensors, k=3, constant=constant)



from mygrad.tensor_base import (  # isort:skip  # avoid an import cycle
    asarray,
    astensor,
    tensor,
    Tensor,
)
from mygrad._dtype_mirrors import *
from mygrad._utils.graph_tracking import no_autodiff
from mygrad._utils.lock_management import (
    mem_guard_active,
    mem_guard_off,
    mem_guard_on,
    turn_memory_guarding_off,
    turn_memory_guarding_on,
)
from mygrad.indexing_routines.funcs import *
from mygrad.linalg.funcs import einsum
from mygrad.math.arithmetic.funcs import *
from mygrad.math.consts import *
from mygrad.math.exp_log.funcs import *
from mygrad.math.hyperbolic_trig.funcs import *
from mygrad.math.misc.funcs import *
from mygrad.math.nondifferentiable import any, argmax, argmin
from mygrad.math.sequential.funcs import *
from mygrad.math.sequential.funcs import max, min
from mygrad.math.trigonometric.funcs import *
from mygrad.nnet.layers.utils import sliding_window_view
from mygrad.no_grad_funcs import *
from mygrad.tensor_creation.funcs import *
from mygrad.tensor_manip.array_shape.funcs import *
from mygrad.tensor_manip.tensor_joining.funcs import *
from mygrad.tensor_manip.tiling.funcs import *
from mygrad.tensor_manip.transpose_like.funcs import *
from mygrad.ufuncs._ufunc_creators import ufunc
from ._io import load, save

from . import random
from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions


setattr(Tensor, "clip", clip)
execute_op = Tensor._op



from typing import Optional, Tuple, Union

from mygrad.tensor_base import Tensor, implements_numpy_override
from mygrad.typing import ArrayLike

from .ops import MoveAxis, Roll, SwapAxes, Transpose

__all__ = ["transpose", "moveaxis", "swapaxes", "roll"]


@implements_numpy_override()
def transpose(a: ArrayLike, *axes: int, constant: Optional[bool] = None) -> Tensor:
    """Permute the dimensions of a tensor.

    Parameters
    ----------
    a : ArrayLike
        The tensor to be transposed

    axes : int
        By default, reverse the dimensions, otherwise permute the axes
        according to the values given.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    mygrad.Tensor
        `a` with its axes permuted.  A new tensor is returned.

    Examples
    --------
    >>> import mygrad as mg
    >>> a = mg.tensor([[1, 2], [3, 4]])
    >>> a
    Tensor([[1, 2],
            [3, 4]])
    >>> a.transpose()
    Tensor([[1, 3],
            [2, 4]])
    >>> a.transpose((1, 0))
    Tensor([[1, 3],
            [2, 4]])
    >>> a.transpose(1, 0)
    Tensor([[1, 3],
            [2, 4]])"""
    if not axes:
        axes = None
    elif hasattr(axes[0], "__iter__") or axes[0] is None:
        if len(axes) > 1:
            raise TypeError(
                f"'{type(axes[0])}' object cannot be interpreted as an integer"
            )
        axes = axes[0]
    return Tensor._op(Transpose, a, op_args=(axes,), constant=constant)


@implements_numpy_override()
def moveaxis(
    a: ArrayLike,
    source: Union[int, Tuple[int, ...]],
    destination: Union[int, Tuple[int, ...]],
    *,
    constant: Optional[bool] = None,
) -> Tensor:
    """Move axes of a tensor to new positions. Other axes remain in their
    original order.


    Parameters
    ----------
    a : ArrayLike
        The array whose axes should be reordered.

    source : Union[int, Sequence[int]]
        Original positions of the axes to move. These must be unique.

    destination : Union[int, Sequence[int]]
        Destination positions for each of the original axes. These must also be
        unique.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.
    Returns
    -------
    result : mygrad.Tensor
        Array with moved axes. This array is a view of the input array..

    Examples
    --------
    >>> from mygrad import zeros, moveaxis
    >>> x = zeros((3, 4, 5))
    >>> moveaxis(x, 0, -1).shape
    (4, 5, 3)
    >>> moveaxis(x, -1, 0).shape
    (5, 3, 4)
    >>> moveaxis(x, [0, 1], [-1, -2]).shape
    (5, 4, 3)"""
    return Tensor._op(MoveAxis, a, op_args=(source, destination), constant=constant)


@implements_numpy_override()
def swapaxes(
    a: ArrayLike, axis1: int, axis2: int, *, constant: Optional[bool] = None
) -> Tensor:
    """Interchange two axes of a tensor.

    Parameters
    ----------
    a : ArrayLike
        Input array.

    axis1 : int
        First axis.

    axis2 : int
        Second axis.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    mygrad.Tensor

    Examples
    --------
    >>> from mygrad import Tensor, swapaxes
    >>> x = Tensor([[1, 2, 3]])
    >>> swapaxes(x, 0, 1)
    Tensor([[1],
           [2],
           [3]])
    >>> x = Tensor([[[0, 1], [2, 3]], [[4, 5], [6, 7]]])
    >>> x
    Tensor([[[0, 1],
            [2, 3]],
           [[4, 5],
            [6, 7]]])
    >>> swapaxes(x, 0, 2)
    Tensor([[[0, 4],
            [2, 6]],
           [[1, 5],
            [3, 7]]])
    """
    return Tensor._op(SwapAxes, a, op_args=(axis1, axis2), constant=constant)


@implements_numpy_override()
def roll(
    a: ArrayLike,
    shift: Union[int, Tuple[int, ...]],
    axis=None,
    *,
    constant: Optional[bool] = None,
) -> Tensor:
    """
    Roll tensor elements along a given axis.

    Elements that roll beyond the end of an axis "wrap back around" to the beginning.

    This docstring was adapted from ``numpy.roll``

    Parameters
    ----------
    a : ArrayLike
        Input tensor.

    shift : Union[int, Tuple[int, ...]]
        The number of places by which elements are shifted.  If a tuple,
        then `axis` must be a tuple of the same size, and each of the
        given axes is shifted by the corresponding number.  If an int
        while `axis` is a tuple of ints, then the same value is used for
        all given axes.

    axis : Optional[Union[int, Tuple[int, ...]]]
        Axis or axes along which elements are shifted.  By default, the
        array is flattened before shifting, after which the original
        shape is restored.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    res : Tensor
        Output array, with the same shape as `a`.

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.arange(10)
    >>> mg.roll(x, 2)
    Tensor([8, 9, 0, 1, 2, 3, 4, 5, 6, 7])
    >>> x2 = mg.reshape(x, (2,5))
    >>> x2
    Tensor([[0, 1, 2, 3, 4],
           [5, 6, 7, 8, 9]])
    >>> mg.roll(x2, 1)
    Tensor([[9, 0, 1, 2, 3],
           [4, 5, 6, 7, 8]])
    >>> mg.roll(x2, 1, axis=0)
    Tensor([[5, 6, 7, 8, 9],
           [0, 1, 2, 3, 4]])
    >>> mg.roll(x2, 1, axis=1)
    Tensor([[4, 0, 1, 2, 3],
           [9, 5, 6, 7, 8]])
    """
    return Tensor._op(
        Roll, a, op_kwargs=dict(shift=shift, axis=axis), constant=constant
    )



import sys
from typing import TYPE_CHECKING, List, Sequence, Tuple, TypeVar, Union

import numpy as np

if TYPE_CHECKING:  # pragma: no cover
    from mygrad import Tensor

if sys.version_info >= (3, 8):  # pragma: no cover
    from typing import Protocol

    HAS_PROTOCOL = True
else:  # pragma: no cover
    try:
        from typing_extensions import Protocol
    except ImportError:
        HAS_PROTOCOL = False
        Protocol = object
    else:
        HAS_PROTOCOL = True


if not TYPE_CHECKING and not HAS_PROTOCOL:  # pragma: no cover

    class ImplementsArray:
        def __array__(self, dtype: None = ...) -> np.ndarray:
            ...


else:  # pragma: no cover

    class ImplementsArray(Protocol):
        def __array__(self, dtype: None = ...) -> np.ndarray:
            ...


Real = Union[int, float]


sr1 = Sequence[Real]
sr2 = Sequence[sr1]
sr3 = Sequence[sr2]
sr4 = Sequence[sr3]

# Sequence[Union[s1, s2]] is *not* valid!
SequenceNDReals = Union[sr1, sr2, sr3, sr4]

# include Tensor and ndarray explicitly in case `ImplementsArray`
# is not protocol

if TYPE_CHECKING:  # pragma: no cover
    ArrayLike = Union[Real, "Tensor", np.ndarray, ImplementsArray, SequenceNDReals]
else:  # pragma: no cover
    ArrayLike = TypeVar(
        "ArrayLike", Real, "Tensor", np.ndarray, ImplementsArray, SequenceNDReals
    )


sb1 = Sequence[bool]
sb2 = Sequence[sb1]
sb3 = Sequence[sb2]
sb4 = Sequence[sb3]

# Sequence[Union[s1, s2]] is *not* valid!
SequenceNDBools = Union[sb1, sb2, sb3, sb4]

if TYPE_CHECKING:  # pragma: no cover
    Mask = Union[ImplementsArray, np.ndarray, "Tensor", bool, SequenceNDBools]
else:  # pragma: no cover
    Mask = TypeVar(
        "Mask",
        bound=Union[ImplementsArray, np.ndarray, "Tensor", bool, SequenceNDBools],
    )


Index = Union[
    int,
    None,
    slice,
    ImplementsArray,
    np.ndarray,
    Sequence[int],
    Tuple["Index"],
    List["Index"],
]



from typing import Optional

import numpy as np

from mygrad import Tensor, asarray, mean
from mygrad.typing import ArrayLike

from ._utils import check_loss_inputs


def negative_log_likelihood(
    x: ArrayLike,
    y_true: ArrayLike,
    *,
    weights: Optional[ArrayLike] = None,
    constant: Optional[bool] = None,
) -> Tensor:
    """Returns the (weighted) negative log-likelihood loss between log-probabilities and y_true.

    Note that this does not compute a softmax, so you should input log-probabilities to this.
    See ``softmax_crossentropy`` if you need your loss to compute a softmax.

    Parameters
    ----------
    x : ArrayLike, shape=(N, C)
        The C log-probabilities for each of the N pieces of data.

    y_true : ArrayLike, shape=(N,)
        The correct class indices, in [0, C), for each datum.

    weights : ArrayLike, shape=(C,) optional (default=None)
        The weighting factor to use on each class, or None.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor, shape=()
        The average (weighted) negative log-likelihood loss.

    Examples
    --------
    >>> import mygrad as mg
    >>> from mygrad.nnet import negative_log_likelihood

    Let's take a simple case where N=1, and C=3. We'll thus make up classification
    scores for a single datum. Suppose the scores are identical for the three classes
    and that the true class is class-0, so that the log-probs are each 1/3:

    >>> logprob = mg.log(1 / 3).item()
    >>> x = mg.Tensor([[logprob, logprob, logprob]])  # a shape-(1, 3) tensor of log-probabilities
    >>> y_true = mg.Tensor([0])  # the correct class for this datum is class-0
    >>> negative_log_likelihood(x, y_true)
    Tensor(1.09861229)

    Log-probabilities where the prediction is highly-confident and correct:

    >>> x = mg.Tensor([[0, -20, -20]])
    >>> negative_log_likelihood(x, y_true)
    Tensor(0.)

    Adding a class-weighting:

    >>> x = mg.Tensor([[-4.6, -4.6, -0.02]])
    >>> weights = mg.Tensor([2, 1, 1])
    >>> negative_log_likelihood(x, y_true, weights=weights)
    Tensor(9.2)
    """
    if isinstance(y_true, Tensor):
        y_true = y_true.data
    check_loss_inputs(x, y_true)

    if weights is None:
        weights = np.ones(x.shape[1])

    weights = asarray(weights)

    if weights.ndim != 1 or weights.shape[0] != x.shape[1]:
        raise ValueError(
            "`weights` must be a shape-(C,) array: \n"
            f"\tExpected shape-{x.shape[1]}\n"
            f"\tGot shape-{y_true.shape}"
        )

    label_locs = (range(len(y_true)), y_true)
    factors = weights[y_true]
    return -mean(x[label_locs] * factors, constant=constant)



from typing import Optional, Sequence, Union

import numpy as np

from mygrad.tensor_base import Tensor, _resolve_constant, implements_numpy_override
from mygrad.typing import ArrayLike, DTypeLikeReals, Real

Shape = Union[Sequence[int], int]


def _anything_but_tensor(x):
    if isinstance(x, Tensor):
        x = x.data
    return x


__all__ = [
    "arange",
    "empty",
    "empty_like",
    "eye",
    "geomspace",
    "identity",
    "linspace",
    "logspace",
    "ones",
    "ones_like",
    "full",
    "full_like",
    "zeros",
    "zeros_like",
]


def empty(
    shape: Shape, dtype: DTypeLikeReals = np.float32, *, constant: Optional[bool] = None
) -> Tensor:
    """Return a new Tensor of the given shape and type, without initializing entries.

    This docstring was adapted from ``numpy.empty`` [1]_

    Parameters
    ----------
    shape : Union[int, Tuple[int]]
        The shape of the empty array.

    dtype : data-type, optional (default=numpy.float32)
        The data type of the output Tensor.

    constant : Optional[bool]
        If ``True``, this tensor is a constant, and thus does not facilitate
        back propagation.

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    Tensor
        A tensor of uninitialized data of the given shape and dtype.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.empty.html

    See Also
    --------
    empty_like : Return an empty tensor with shape and type of input.
    ones : Return a new tensor setting values to one.
    zeros : Return a new tensor setting values to zero.
    full : Return a new tensor of given shape filled with value.


    Notes
    -----
    `empty`, unlike `zeros`, does not set the array values to zero,
    and may therefore be marginally faster.  On the other hand, it requires
    the user to manually set all the values in the array, and should be
    used with caution.

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.empty([2, 2], constant=True)
    Tensor([[ -9.74499359e+001,   6.69583040e-309],
            [  2.13182611e-314,   3.06959433e-309]])         #random

    >>> mg.empty([2, 2], dtype=int)
    Tensor([[-1073741821, -1067949133],
            [  496041986,    19249760]])                     #random
    """
    return Tensor(np.empty(shape=shape, dtype=dtype), constant=constant, copy=False)


@implements_numpy_override()
def empty_like(
    other: ArrayLike,
    dtype: Optional[DTypeLikeReals] = None,
    shape: Optional[Union[int, Sequence[int]]] = None,
    *,
    constant: Optional[bool] = None,
) -> Tensor:
    """Return a new Tensor of the same shape and type as the given array.

    This docstring was adapted from ``numpy.empty_like`` [1]_

    Parameters
    ----------
    other : ArrayLike
        The Tensor or array whose shape and datatype should be mirrored.

    dtype : Optional[DTypeLikeReals]
        Override the data type of the returned Tensor with this value, or None to not override.

    shape : Optional[Union[int, Sequence[int]]]
        If specified, overrides the shape of the result

    constant : Optional[bool]
        If ``True``, this tensor is a constant, and thus does not facilitate
        back propagation. If ``None`` then:

        Inferred from ``other``, if other is a tensor
        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

    Returns
    -------
    Tensor
        A tensor of uninitialized data whose shape and type match `other`.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.empty_like.html

    See Also
    --------
    empty : Return a new Tensor of the given shape and type, without initializing entries.
    ones : Return a new tensor setting values to one.
    zeros : Return a new tensor setting values to zero.
    full : Return a new tensor of given shape filled with value.

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.arange(4).reshape(2, 2)
    >>> mg.empty_like(x, constant=True)
    Tensor([[ -9.74499359e+001,   6.69583040e-309],
            [  2.13182611e-314,   3.06959433e-309]])         #random

    >>> mg.empty_like(x, dtype=int)
    Tensor([[-1073741821, -1067949133],
            [  496041986,    19249760]])                     #random
    """
    constant = _resolve_constant(other, constant=constant)
    return Tensor(
        np.empty_like(_anything_but_tensor(other), dtype=dtype, shape=shape),
        constant=constant,
        copy=False,
    )


def eye(
    N: int,
    M: Optional[int] = None,
    k: int = 0,
    dtype: DTypeLikeReals = float,
    *,
    constant: Optional[bool] = None,
) -> Tensor:
    """Return a 2D Tensor with ones on the diagonal and zeros elsewhere.

    This docstring was adapted from ``numpy.eye`` [1]_

    Parameters
    ----------
    N : int
        The number of rows in the output Tensor.

    M : int, optional (default=None)
        The number of columns in the output, or None to match `rows`.

    k : int, optional (default=0)
        The index of the diagonal. 0 is the main diagonal; a positive value is the upper
        diagonal, while a negative value refers to the lower diagonal.

    dtype : data-type, optional (default=numpy.float32)
        The data type of the output Tensor.

    constant : Optional[bool]
        If ``True``, this tensor is a constant, and thus does not facilitate
        back propagation.

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.eye.html

    Returns
    -------
    Tensor
        A tensor whose elements are 0, except for the :math:`k`-th diagonal, whose values are 1.

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.eye(2, dtype=int)
    Tensor([[1, 0],
            [0, 1]])
    >>> mg.eye(3, k=1)
    Tensor([[ 0.,  1.,  0.],
            [ 0.,  0.,  1.],
            [ 0.,  0.,  0.]])
    """
    return Tensor(
        np.eye(N, M=M, k=k, dtype=dtype),
        constant=constant,
        copy=False,
    )


def identity(
    n: int, dtype: DTypeLikeReals = float, *, constant: Optional[bool] = None
) -> Tensor:
    """Return the identity Tensor; a square Tensor with 1s on the main diagonal and 0s elsewhere.

    This docstring was adapted from ``numpy.identity`` [1]_

    Parameters
    ----------
    n : int
        The number of rows and columns in the output Tensor.

    dtype : data-type, optional (default=numpy.float32)
        The data type of the output Tensor.

    constant : Optional[bool]
        If ``True``, this tensor is a constant, and thus does not facilitate
        back propagation.

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    Tensor
        A square Tensor whose main diagonal is 1 and all other elements are 0.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.identity.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.identity(3)
    Tensor([[ 1.,  0.,  0.],
            [ 0.,  1.,  0.],
            [ 0.,  0.,  1.]])
    """
    return Tensor(np.identity(n, dtype=dtype), constant=constant, copy=False)


def ones(
    shape: Shape, dtype: DTypeLikeReals = np.float32, *, constant: Optional[bool] = None
) -> Tensor:
    """
    Return a Tensor of the given shape and type, filled with ones.

    This docstring was adapted from ``numpy.ones`` [1]_

    Parameters
    ----------
    shape : Union[int, Tuple[int]]
        The shape of the output Tensor.

    dtype : data-type, optional (default=numpy.float32)
        The data type of the output Tensor.

    constant : Optional[bool]
        If ``True``, this tensor is a constant, and thus does not facilitate
        back propagation.

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    Tensor
        A Tensor of ones with the given shape and data type.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.ones.html

    See Also
    --------
    ones_like : Return an tensor of ones with shape and type of input.
    empty : Return a new uninitialized tensor.
    zeros : Return a new tensor setting values to zero.
    full : Return a new tensor of given shape filled with value.

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.ones(5)
    Tensor([ 1.,  1.,  1.,  1.,  1.])

    >>> mg.ones((5,), dtype=int)
    Tensor([1, 1, 1, 1, 1])

    >>> mg.ones((2, 1))
    Tensor([[ 1.],
           [ 1.]])

    >>> mg.ones((2, 2))
    Tensor([[ 1.,  1.],
            [ 1.,  1.]])
    """
    return Tensor(np.ones(shape, dtype=dtype), constant=constant, copy=False)


@implements_numpy_override()
def ones_like(
    other: ArrayLike,
    dtype: Optional[DTypeLikeReals] = None,
    shape: Optional[Union[int, Sequence[int]]] = None,
    *,
    constant: Optional[bool] = None,
) -> Tensor:
    """
    Return a Tensor of the same shape and type as the given, filled with ones.

    This docstring was adapted from ``numpy.ones_like`` [1]_

    Parameters
    ----------
    other : array_like
        The Tensor or array whose shape and datatype should be mirrored.

    dtype : Optional[DTypeLikeReals]
        Override the data type of the returned Tensor with this value, or None to not override.

    shape : Optional[Union[int, Sequence[int]]]
        If specified, overrides the shape of the result

    constant : Optional[bool]
        If ``True``, this tensor is a constant, and thus does not facilitate
        back propagation. If ``None`` then:

        Inferred from ``other``, if other is a tensor
        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.


    Returns
    -------
    Tensor
        A Tensor of ones whose shape and data type match `other`.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.ones_like.html

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.arange(6).reshape((2, 3))
    >>> x
    Tensor([[0, 1, 2],
            [3, 4, 5]])

    >>> mg.ones_like(x)
    Tensor([[1, 1, 1],
            [1, 1, 1]])

    >>> y = mg.arange(3, dtype=float)
    >>> y
    Tensor([ 0.,  1.,  2.])

    >>> mg.ones_like(y)
    Tensor([ 1.,  1.,  1.])
    """
    constant = _resolve_constant(other, constant=constant)

    return Tensor(
        np.ones_like(_anything_but_tensor(other), dtype=dtype, shape=shape),
        constant=constant,
        copy=False,
    )


def zeros(
    shape: Shape, dtype: DTypeLikeReals = np.float32, *, constant: Optional[bool] = None
) -> Tensor:
    """
    Return a Tensor of the given shape and type, filled with zeros.

    This docstring was adapted from ``numpy.zeros`` [1]_

    Parameters
    ----------
    shape : Union[int, Tuple[int]]
        The shape of the output Tensor.

    dtype : data-type, optional (default=numpy.float32)
        The data type of the output Tensor.

    constant : Optional[bool]
        If ``True``, this tensor is a constant, and thus does not facilitate
        back propagation.

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    Tensor
        A Tensor of zeros with the given shape and data type.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.zeros.html

    See Also
    --------
    ones_like : Return an tensor of ones with shape and type of input.
    empty : Return a new uninitialized tensor.
    ones : Return a new tensor setting values to one.
    full : Return a new tensor of given shape filled with value.

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.zeros(5)
    Tensor([ 0.,  0.,  0.,  0.,  0.])

    >>> mg.zeros((5,), dtype=int, constant=True) # tensor will not back-propagate a gradient
    Tensor([0, 0, 0, 0, 0])

    >>> mg.zeros((2, 1))
    Tensor([[ 0.],
            [ 0.]])

    >>> mg.zeros((2, 2))
    Tensor([[ 0.,  0.],
            [ 0.,  0.]])
    """
    return Tensor(np.zeros(shape, dtype), constant=constant, copy=False)


@implements_numpy_override()
def zeros_like(
    other: ArrayLike,
    dtype: Optional[DTypeLikeReals] = None,
    shape: Optional[Union[int, Shape]] = None,
    *,
    constant: Optional[bool] = None,
) -> Tensor:
    """
    Return a Tensor of the same shape and type as the given, filled with zeros.

    This docstring was adapted from ``numpy.zeros_like`` [1]_

    Parameters
    ----------
    other : ArrayLike
        The Tensor or array whose shape and datatype should be mirrored.

    dtype : Optional[DTypeLikeReals]
        Override the data type of the returned Tensor with this value, or None to not override.

    shape : Optional[int, Sequence[int]]
        If specified, overrides the shape of the result

    constant : Optional[bool]
        If ``True``, this tensor is a constant, and thus does not facilitate
        back propagation. If ``None`` then:

        Inferred from ``other``, if other is a tensor
        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    Tensor
        A Tensor of zeros whose shape and data type match `other`.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.zeros_like.html

    See Also
    --------
    empty_like : Return an empty tensor with shape and type of input.
    ones_like : Return an tensor of ones with shape and type of input.
    full_like : Return a new tensor with shape of input filled with value.
    zeros : Return a new tensor setting values to zero.

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.arange(6).reshape((2, 3))
    >>> x
    Tensor([[0, 1, 2],
            [3, 4, 5]])

    >>> mg.zeros_like(x, constant=True)  # tensor will not back-propagate a gradient
    Tensor([[0, 0, 0],
            [0, 0, 0]])

    >>> y = mg.arange(3, dtype=float)
    >>> y
    Tensor([ 0.,  1.,  2.])

    >>> mg.zeros_like(y)
    Tensor([ 0.,  0.,  0.])
    """
    constant = _resolve_constant(other, constant=constant)
    return Tensor(
        np.zeros_like(_anything_but_tensor(other), dtype=dtype, shape=shape),
        constant=constant,
        copy=False,
    )


def full(
    shape: Shape,
    fill_value: ArrayLike,
    dtype: Optional[DTypeLikeReals] = None,
    *,
    constant: Optional[bool] = None,
) -> Tensor:
    """
    Return a Tensor of the given shape and type, filled with `fill_value`.

    This docstring was adapted from ``numpy.full`` [1]_

    Parameters
    ----------
    shape : Union[int, Iterable[int]]
        The shape of the output Tensor.

    fill_value : ArrayLike
        The value with which to fill the output Tensor. Note that this function
        is not differentiable – the resulting tensor will not backprop through
        `fill_value`.

        The value with which to fill the output Tensor.

    dtype : Optional[DTypeLikeReals]
        The data type of the output Tensor, or None to match `fill_value`..

    constant : Optional[bool]
        If ``True``, this tensor is a constant, and thus does not facilitate
        back propagation.

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    Tensor
        A Tensor of `fill_value` with the given shape and dtype.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.full.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.full((2, 2), 33)
    Tensor([[ 33,  33],
            [ 33,  33]])

    >>> mg.full((2, 2), 10)
    Tensor([[10, 10],
            [10, 10]])
    """
    return Tensor(
        np.full(shape, fill_value=fill_value, dtype=dtype),
        constant=constant,
        copy=False,
    )


@implements_numpy_override()
def full_like(
    other: ArrayLike,
    fill_value: Real,
    dtype: Optional[DTypeLikeReals] = None,
    shape: Optional[Union[int, Shape]] = None,
    constant: Optional[bool] = None,
) -> Tensor:
    """Return a Tensor of the same shape and type as the given, filled with `fill_value`.

    This docstring was adapted from ``numpy.full_like`` [1]_

    Parameters
    ----------
    other : ArrayLike
        The tensor or array whose shape and datatype should be mirrored.

    fill_value : Real
        The value with which to fill the output Tensor.

    dtype : Optional[DTypeLikeReals]
        Override the data type of the returned Tensor with this value, or None to not override.

    shape : Optional[int, Sequence[int]]
        If specified, overrides the shape of the result

    constant : Optional[bool]
        If ``True``, this tensor is a constant, and thus does not facilitate
        back propagation. If ``None`` then:

        Inferred from ``other``, if other is a tensor
        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

    Returns
    -------
    Tensor
        A Tensor of `fill_value` whose shape and data type match `other`.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.full_like.html

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.arange(6, dtype=int)
    >>> mg.full_like(x, 1)
    Tensor([1, 1, 1, 1, 1, 1])
    >>> mg.full_like(x, 0.1)
    Tensor([0, 0, 0, 0, 0, 0])
    >>> mg.full_like(x, 0.1, dtype=np.double)
    Tensor([ 0.1,  0.1,  0.1,  0.1,  0.1,  0.1])
    >>> mg.full_like(x, np.nan, dtype=np.double)
    Tensor([ nan,  nan,  nan,  nan,  nan,  nan])

    >>> y = mg.arange(6, dtype=np.double)
    >>> mg.full_like(y, 0.1)
    Tensor([ 0.1,  0.1,  0.1,  0.1,  0.1,  0.1])
    """
    constant = _resolve_constant(other, constant=constant)

    return Tensor(
        np.full_like(
            _anything_but_tensor(other),
            fill_value=_anything_but_tensor(fill_value),
            dtype=dtype,
            shape=shape,
        ),
        constant=constant,
        copy=False,
    )


def arange(
    *args,
    constant: Optional[bool] = None,
    **kwargs,
) -> Tensor:
    """
    arange([start,] stop[, step,], dtype=None, *, constant=None)

    Return a Tensor with evenly-spaced values within a given interval.

    Values are generated within [start, stop). Note that for non-integer steps, results may be
    inconsistent; you are better off using `linspace` instead.

    This docstring was adapted from ``numpy.arange`` [1]_

    Parameters
    ----------
    start : Real, optional, default=0
        The start of the interval, inclusive.

    stop : Real
        The end of the interval, exclusive.

    step : int, optional (default=1)
        The spacing between successive values.

    dtype : Optional[DTypeLikeReals]
        The data type of the output Tensor, or None to infer from the inputs.

    constant : Optional[bool]
        If ``True``, this tensor is a constant, and thus does not facilitate
        back propagation.

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    Tensor
        A Tensor of evenly-spaced values in [start, end).

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.arange.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.arange(3)
    Tensor([0, 1, 2])
    >>> mg.arange(3.0, constant=True)  # resulting tensor will not back-propagate a gradient
    Tensor([ 0.,  1.,  2.])
    >>> mg.arange(3,7)
    Tensor([3, 4, 5, 6])
    >>> mg.arange(3,7,2)
    Tensor([3, 5])
    """
    return Tensor(np.arange(*args, **kwargs), constant=constant, copy=False)


def linspace(
    start: ArrayLike,
    stop: ArrayLike,
    num: int = 50,
    endpoint: bool = True,
    dtype: Optional[DTypeLikeReals] = None,
    axis: int = 0,
    *,
    constant: Optional[bool] = None,
) -> Tensor:
    """Return a Tensor with evenly-spaced numbers over a specified interval.

    Values are generated within [start, stop], with the endpoint optionally excluded.

    This docstring was adapted from ``numpy.linspace`` [1]_

    Parameters
    ----------
    start : ArrayLike
        The starting value of the sequence, inclusive.

    stop : ArrayLike
        The ending value of the sequence, inclusive unless `include_endpoint` is False.

    num : int, optional (default=50)
        The number of values to generate. Must be non-negative.

    endpoint : bool, optional (default=True)
        Whether to include the endpoint in the Tensor. Note that if False, the step size changes
        to accommodate the sequence excluding the endpoint.

    dtype : Optional[DTypeLikeReals]
        The data type of the output Tensor, or None to infer from the inputs.

    axis : int, optional (default=0)
        The axis in the result to store the samples - for array-like start/stop.

    constant : Optional[bool]
        If ``True``, this tensor is a constant, and thus does not facilitate
        back propagation.

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    Tensor

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.linspace.html

    See Also
    --------
    arange : Similar to `linspace`, but uses a step size (instead of the
             number of samples).
    logspace : Samples uniformly distributed in log space.

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.linspace(2.0, 3.0, num=5)
    Tensor([ 2.  ,  2.25,  2.5 ,  2.75,  3.  ])
    >>> mg.linspace(2.0, 3.0, num=5, endpoint=False)
    Tensor([ 2. ,  2.2,  2.4,  2.6,  2.8])
    """
    return Tensor(
        np.linspace(
            start,
            stop,
            num,
            endpoint=endpoint,
            dtype=dtype,
            axis=axis,
        ),
        constant=constant,
        copy=False,
    )


def logspace(
    start: ArrayLike,
    stop: ArrayLike,
    num: int = 50,
    endpoint: bool = True,
    base: Real = 10,
    dtype: Optional[DTypeLikeReals] = None,
    axis: int = 0,
    *,
    constant: Optional[bool] = None,
) -> Tensor:
    """Return a Tensor with evenly-spaced numbers over a specified interval on a log scale.
    This is not a differentiable function - it does not propagate gradients to its inputs.

    In linear space, values are generated within [base**start, base**stop], with the endpoint
    optionally excluded.

    This docstring was adapted from ``numpy.logspace`` [1]_

    Parameters
    ----------
    start : ArrayLike
        The starting value of the sequence, inclusive; start at `base ** start`.

    stop : ArrayLike
        The ending value of the sequence, inclusive unless `include_endpoint` is False; end at
        `base ** stop`.

    num : int, optional (default=50)
        The number of values to generate. Must be non-negative.

    endpoint : bool, optional (default=True)
        Whether to include the endpoint in the Tensor. Note that if False, the step size changes
        to accommodate the sequence excluding the endpoint.

    base : Real, optional (default=10)
        The base of the log space.

    dtype : Optional[DTypeLikeReals]
        The data type of the output Tensor, or None to infer from the inputs.

    axis : int, optional (default=0)
        The axis in the result to store the samples - for array-like start/stop.

    constant : Optional[bool]
        If ``True``, this tensor is a constant, and thus does not facilitate
        back propagation.

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    Tensor

    See Also
    --------
    arange : Similar to linspace, with the step size specified instead of the
             number of samples. Note that, when used with a float endpoint, the
             endpoint may or may not be included.
    linspace : Similar to logspace, but with the samples uniformly distributed
               in linear space, instead of log space.
    geomspace : Similar to logspace, but with endpoints specified directly.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.logspace.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.logspace(2.0, 3.0, num=4)
    Tensor([  100.        ,   215.443469  ,   464.15888336,  1000.        ])
    >>> mg.logspace(2.0, 3.0, num=4, endpoint=False)
    Tensor([ 100.        ,  177.827941  ,  316.22776602,  562.34132519])
    >>> mg.logspace(2.0, 3.0, num=4, base=2.0)
    Tensor([ 4.        ,  5.0396842 ,  6.34960421,  8.        ])

    """
    return Tensor(
        np.logspace(
            start=start,
            stop=stop,
            num=num,
            endpoint=endpoint,
            base=base,
            dtype=dtype,
            axis=axis,
        ),
        constant=constant,
        copy=False,
    )


def geomspace(
    start: ArrayLike,
    stop: ArrayLike,
    num=50,
    endpoint=True,
    dtype=None,
    axis=0,
    *,
    constant: Optional[bool] = None,
) -> Tensor:
    """Return a Tensor with evenly-spaced values in a geometric progression.

    Each output sample is a constant multiple of the previous output.

    This docstring was adapted from ``numpy.geomspace`` [1]_

    Parameters
    ----------
    start : ArrayLike
        The starting value of the output.

    stop : ArrayLike
        The ending value of the sequence, inclusive unless `endpoint` is false.

    num : int, optional (default=50)
        The number of values to generate. Must be non-negative.

    endpoint : bool, optional (default=True)
        Whether to include the endpoint in the Tensor. Note that if False, the step size changes
        to accommodate the sequence excluding the endpoint.

    dtype : Optional[DTypeLikeReals]
        The data type of the output Tensor, or None to infer from the inputs.

    axis : int, optional (default=0)
        The axis in the result to store the samples - for array-like start/stop.

    constant : Optional[bool]
        If ``True``, this tensor is a constant, and thus does not facilitate
        back propagation.

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    Tensor

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.geomspace.html

    See Also
    --------
    logspace : Similar to geomspace, but with endpoints specified using log
               and base.
    linspace : Similar to geomspace, but with arithmetic instead of geometric
               progression.
    arange : Similar to linspace, with the step size specified instead of the
             number of samples.

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.geomspace(1, 1000, num=4)
    Tensor([    1.,    10.,   100.,  1000.])
    >>> mg.geomspace(1, 1000, num=3, endpoint=False)
    Tensor([   1.,   10.,  100.])
    >>> mg.geomspace(1, 1000, num=4, endpoint=False)
    Tensor([   1.        ,    5.62341325,   31.6227766 ,  177.827941  ])
    >>> mg.geomspace(1, 256, num=9)
    Tensor([   1.,    2.,    4.,    8.,   16.,   32.,   64.,  128.,  256.])

    Note that the above may not produce exact integers:

    >>> mg.geomspace(1, 256, num=9, dtype=int)
    Tensor([  1,   2,   4,   7,  16,  32,  63, 127, 256])
    >>> np.around(mg.geomspace(1, 256, num=9).data).astype(int)
    array([  1,   2,   4,   8,  16,  32,  64, 128, 256])

    Negative, and decreasing inputs are allowed:

    >>> mg.geomspace(1000, 1, num=4)
    Tensor([ 1000.,   100.,    10.,     1.])
    >>> mg.geomspace(-1000, -1, num=4)
    Tensor([-1000.,  -100.,   -10.,    -1.])
    """
    return Tensor(
        np.geomspace(
            start=start,
            stop=stop,
            num=num,
            endpoint=endpoint,
            dtype=dtype,
            axis=axis,
        ),
        constant=constant,
        copy=False,
    )



from typing import Optional

import numpy as np

from mygrad import Tensor


def dirac(*shape: int, dtype=np.float32, constant: Optional[bool] = None) -> Tensor:
    """Initialize a `Tensor` according to the Dirac initialization procedure described by Zagoruyko and Komodakis.

    Parameters
    ----------
    shape : Sequence[int]
        The shape of the output Tensor. Note that `shape` must be at least two-dimensional.

    dtype : data-type, optional (default=float32)
        The data type of the output tensor.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    mygrad.Tensor, shape=`shape`
        A Tensor, with values initialized according to the Dirac initialization.

    Extended Description
    --------------------
    Zagoruyko and Komodakis put forward the Dirac initialization in the paper
        "DiracNets: Training Very Deep Neural Networks without Skip Connections"
    https://arxiv.org/abs/1706.00388

    A Tensor I initialized via this should satisfy:
        I ⋆ x = x

    for compatible tensors x, where ⋆ indicates convolution. Note that this does not
    guarantee that the convolution will produce x, but it will preserve as many channels of
    the input as possible.
    """
    if len(shape) == 1:
        shape = shape[0]

    if len(shape) < 2:
        raise ValueError("Dirac initialization requires at least two dimensions")

    tensor = np.zeros(shape, dtype=dtype)
    minimum_depth = np.minimum(shape[0], shape[1])  # out dim, in dim
    depths = range(minimum_depth)
    trailing_indices = ([i // 2] * len(depths) for i in tensor.shape[2:])
    # tensor[i, i, k1//2, k2//2, ..., kn//2] for each i in min(shape[0], shape[1]
    # where the k values are the spatial dimensions of `tensor`
    tensor[(depths, depths, *trailing_indices)] = 1
    return Tensor(tensor, constant=constant)



import numpy as np

__all__ = [
    "Inf",
    "Infinity",
    "NAN",
    "NINF",
    "NZERO",
    "NaN",
    "PINF",
    "PZERO",
    "e",
    "euler_gamma",
    "inf",
    "infty",
    "nan",
    "newaxis",
    "pi",
]

Inf = np.Inf
Infinity = np.Infinity
NAN = np.NAN
NINF = np.NINF
NZERO = np.NZERO
NaN = np.NaN
PINF = np.PINF
PZERO = np.PZERO
e = np.e
euler_gamma = np.euler_gamma
inf = np.inf
infty = np.infty
nan = np.nan
newaxis = np.newaxis
pi = float(np.pi)



from typing import Optional, Tuple, Union

import numpy as np

from mygrad import Tensor
from mygrad.operation_base import Operation
from mygrad.typing import ArrayLike

__all__ = ["batchnorm"]


# TODO: Remove affine parameters from Operation
class BatchNorm(Operation):
    """
    Attributes
    ----------
    mean : numpy.ndarray
    var : numpy.ndarray

    Notes
    -----
    `mean` and `var` are bound as instance-attributes upon
    calling the batch-norm instance.
    """

    def __call__(self, x, gamma, beta, *, eps):
        """
        y(x) = (x - E[x]) / sqrt(Var[x} + eps)
        batchnorm(x) = gamma * y(x) + beta

        Parameters
        ----------
        x : mygrad.Tensor
        gamma : Optional[mygrad.Tensor]
        beta : Optional[mygrad.Tensor]
        eps : Real
           A small non-negative number.

        Returns
        -------
        numpy.ndarray
        """
        normed_dims = tuple(i for i in range(x.ndim) if i != 1)
        keepdims_shape = tuple(1 if n != 1 else d for n, d in enumerate(x.shape))

        self.variables = tuple(i for i in (x, gamma, beta))

        if gamma.size == 0:
            gamma = None
        if beta.size == 0:
            beta = None

        self.gamma = gamma
        self.beta = beta

        x = x.data
        self.x_norm = None  # required for backprop through gamma
        self.mean = x.mean(axis=normed_dims)
        self.var = x.var(axis=normed_dims)

        if eps:
            self.var += eps

        y = x - self.mean.reshape(keepdims_shape)
        self._std = np.sqrt(self.var).reshape(keepdims_shape)  # sqrt(var + eps)
        y /= self._std
        self.x_norm = y
        # optional affine transformation
        if gamma is not None:
            gamma = gamma.data
            # must copy `y` to prevent mutation of `self.x_norm`
            y = y * gamma.reshape(keepdims_shape)

        if beta is not None:
            beta = beta.data
            y = y + beta.reshape(keepdims_shape)
        return y

    def backward_var(self, grad, index, **kwargs):
        x = self.variables[0].data
        if index == 0:  # backprop through x
            normed_dims = tuple(i for i in range(x.ndim) if i != 1)
            keepdims_shape = tuple(1 if n != 1 else d for n, d in enumerate(x.shape))
            N = x.size / x.shape[1]

            # all sums carried over non-channel dims
            # (1/sqrt(var + eps)) * [dL - dL.mean() - (1/N)*x_norm*(x_norm @ dL)]
            grad_ = grad - np.mean(grad, axis=normed_dims, keepdims=True)

            rterm = self.x_norm * np.reshape(
                np.einsum(grad, range(x.ndim), self.x_norm, range(x.ndim), [1]),
                keepdims_shape,
            )
            rterm /= N
            grad_ -= rterm
            grad_ /= self._std
            if (
                self.gamma is not None
            ):  # backprop through optional affine transformation
                gamma = self.gamma.data
                grad_ *= gamma.reshape(keepdims_shape)
            return grad_

        elif index == 1 and self.gamma is not None:  # backprop through gamma
            return np.einsum(grad, range(x.ndim), self.x_norm, range(x.ndim), [1])

        elif (index == 1 and self.gamma is None) or index == 2:
            normed_dims = tuple(i for i in range(x.ndim) if i != 1)
            return grad.sum(axis=normed_dims)
        else:  # pragma: no cover
            raise IndexError


def batchnorm(
    x: ArrayLike,
    *,
    gamma: Optional[ArrayLike] = None,
    beta: Optional[ArrayLike] = None,
    eps: float,
    constant: Optional[bool] = None
) -> Tensor:
    """
    Performs batch normalization on ``x``::

                 y(x) = (x - E[x]) / sqrt(Var[x] + eps)
                 batchnorm(x) = gamma * y(x) + beta

    Where :math:`E[x]` and :math:`Var[x]` represent the mean and variance, respectively,
    over axis-1 of ``x``. The subsequent affine transformation on ``y``
    is optional.

    Parameters
    ----------
    x : array_like, shape=(N, C, ...)
        The batch to be normalized within each entry of C

    gamma : Optional[array_like], shape=(C,)
        Optional per-channel scaling factors to be applied after the
        normalization step.

    beta  : Optional[array_like], shape=(C,)
        Optional per-channel scaling bias factors to be applied after the
        normalization step.

    eps : Real
       A small non-negative number.

    constant : bool, optional (default=False)
        If True, the resulting Tensor is a constant.

    Returns
    -------
    mygrad.Tensor
        The batch-normalized data.

    Examples
    --------
    >>> import mygrad as mg
    >>> from mygrad.nnet import batchnorm
    >>> x = mg.Tensor([1., 4., 1.]).reshape(3, 1)
    >>> batchnorm(x, eps=0)
    Tensor([[-0.70710678],
            [ 1.41421356],
            [-0.70710678]])
    """
    # pass gamma and beta as empty arrays if they are not supplied
    if gamma is None:
        gamma = np.array([])
    if beta is None:
        beta = np.array([])
    return Tensor._op(
        BatchNorm, x, gamma, beta, op_kwargs=dict(eps=eps), constant=constant
    )



import itertools
import sys
from typing import TYPE_CHECKING, Optional, Tuple, Union

import numpy

from mygrad.tensor_base import (
    _REGISTERED_BOOL_ONLY_UFUNC,
    _REGISTERED_CONST_ONLY_UFUNC,
    _REGISTERED_NO_DIFF_NUMPY_FUNCS,
)
from mygrad.typing import ArrayLike, DTypeLike, Mask

__all__ = [
    "allclose",
    "bincount",
    "can_cast",
    "ceil",
    "copyto",
    "divmod",
    "equal",
    "floor",
    "floor_divide",
    "fmod",
    "greater",
    "greater_equal",
    "isclose",
    "isfinite",
    "isinf",
    "isnan",
    "isnat",
    "less",
    "less_equal",
    "logical_and",
    "logical_not",
    "logical_or",
    "logical_xor",
    "may_share_memory",
    "min_scalar_type",
    "mod",
    "not_equal",
    "remainder",
    "result_type",
    "rint",
    "shape",
    "shares_memory",
    "sign",
    "signbit",
    "trunc",
]

_module = sys.modules[__name__]


for _func in itertools.chain.from_iterable(
    (
        _REGISTERED_NO_DIFF_NUMPY_FUNCS,
        _REGISTERED_BOOL_ONLY_UFUNC,
        _REGISTERED_CONST_ONLY_UFUNC,
    )
):
    setattr(_module, _func.__name__, _func)

mod = numpy.remainder

if TYPE_CHECKING:  # pragma: no cover

    def allclose(
        a: ArrayLike,
        b: ArrayLike,
        rtol: float = 1e-05,
        atol: float = 1e-08,
        equal_nan: bool = False,
    ) -> bool:
        pass

    def isclose(
        a: ArrayLike,
        b: ArrayLike,
        rtol: float = 1e-05,
        atol: float = 1e-08,
        equal_nan: bool = False,
    ) -> bool:
        pass

    def bincount(
        x: ArrayLike, weights: Optional[ArrayLike] = None, minlength: int = 0
    ) -> numpy.ndarray:
        pass

    def can_cast(from_, to, casting="safe") -> bool:
        pass

    def copyto(
        dst: ArrayLike, src: ArrayLike, casting: str = "same_kind", where: bool = True
    ):
        pass

    def may_share_memory(
        a: ArrayLike, b: ArrayLike, max_work: Optional[int] = None
    ) -> bool:
        pass

    def min_scalar_type(a: ArrayLike) -> numpy.dtype:
        pass

    def result_type(*arrays_and_dtypes: Union[ArrayLike, numpy.dtype]) -> numpy.dtype:
        pass

    def shares_memory(
        a: ArrayLike, b: ArrayLike, max_work: Optional[int] = None
    ) -> bool:
        pass

    def shape(a: ArrayLike) -> Tuple[int, ...]:
        pass

    def isnan(
        x: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def isfinite(
        x: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def isinf(
        x: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def isnat(
        x: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def signbit(
        x: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def logical_not(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def logical_and(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def logical_or(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def logical_xor(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def greater(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def greater_equal(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def less(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def less_equal(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def equal(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def not_equal(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Union[numpy.ndarray, bool]:
        ...

    def floor_divide(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> numpy.ndarray:
        ...

    def remainder(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> numpy.ndarray:
        ...

    def divmod(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        ...

    def mod(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        ...

    def fmod(
        x1: ArrayLike,
        x2: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> Tuple[numpy.ndarray, numpy.ndarray]:
        ...

    def rint(
        x: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> numpy.ndarray:
        ...

    def sign(
        x: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> numpy.ndarray:
        ...

    def floor(
        x: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> numpy.ndarray:
        ...

    def ceil(
        x: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> numpy.ndarray:
        ...

    def trunc(
        x: ArrayLike,
        out: Optional[ArrayLike] = None,
        *,
        casting: str = "same_kind",
        where: Mask = True,
        order: str = "K",
        dtype: DTypeLike = None,
        subok: bool = True
    ) -> numpy.ndarray:
        ...



from pathlib import Path
from typing import BinaryIO, Union

import numpy as np

import mygrad.tensor_base as tb

_FileLike = Union[str, Path, BinaryIO]


def save(file: _FileLike, tensor: tb.Tensor) -> None:
    """Saves a tensor and its gradient information.

    This docstring was adapted from that of numpy.save()

    Parameters
    ----------
    file : str | Path | BinaryIO
        The file or file-path that where the tensor data and its gradient
        will be saved. Note that the file will be saved as a .npz file.

    tensor : Tensor
        The tensor to be saved. If it has an associated gradient, that will
        be saved as well.

    Notes
    -----
    This function uses ``numpy.savez(file, data=tensor.data, grad=tensor.grad)``
    to save the tensor's data and its gradient. No ``grad`` field is included
    if the tensor does not have a gradient.

    See Also
    --------
    mygrad.load

    Examples
    --------
    >>> import mygrad as mg
    >>> from tempfile import TemporaryFile
    >>> outfile = TemporaryFile()
    >>> x = mg.arange(10.0)
    >>> mg.save(outfile, x)
    >>> _ = outfile.seek(0) # Only needed here to simulate closing & reopening file
    >>> mg.load(outfile)
    Tensor([0., 1., 2., 3., 4., 5., 6., 7., 8., 9.])

    An example of saving a tensor that has an associated gradient.

    >>> (x * x).backward()
    >>> x.grad
    array([ 0.,  2.,  4.,  6.,  8., 10., 12., 14., 16., 18.])
    >>> outfile = TemporaryFile()
    >>> x = mg.arange(10.0)
    >>> mg.save(outfile, x)
    >>> _ = outfile.seek(0) # Only needed here to simulate closing & reopening file
    >>> loaded = mg.load(outfile)
    >>> loaded
    Tensor([0., 1., 2., 3., 4., 5., 6., 7., 8., 9.])
    >>> loaded.grad
    array([ 0.,  2.,  4.,  6.,  8., 10., 12., 14., 16., 18.])
    """
    if not isinstance(tensor, tb.Tensor):
        raise TypeError(
            f"mygrad.save requires a Tensor-type object, got type {type(tensor)}"
        )

    if tensor.grad is not None:
        np.savez(file, data=tensor.data, grad=tensor.grad)
    else:
        np.savez(file, data=tensor.data)


def load(file: _FileLike) -> tb.Tensor:
    """Loads a saved Tensor and its gradient information (if applicable).

    This docstring was adapted from that of numpy.load()

    Parameters
    ----------
    file : str | Path | BinaryIO
        The name of the file that holds the tensor data to load.

    Returns
    -------
    loaded : Tensor
        The loaded tensor (whose gradient will be loaded if it was saved).

    See Also
    --------
    mygrad.save

    Examples
    --------
    >>> import mygrad as mg
    >>> from tempfile import TemporaryFile
    >>> outfile = TemporaryFile()
    >>> x = mg.arange(10.0)
    >>> mg.save(outfile, x)
    >>> _ = outfile.seek(0) # Only needed here to simulate closing & reopening file
    >>> mg.load(outfile)
    Tensor([0., 1., 2., 3., 4., 5., 6., 7., 8., 9.])

    An example of saving a tensor that has an associated gradient.

    >>> (x * x).backward()
    >>> x.grad
    array([ 0.,  2.,  4.,  6.,  8., 10., 12., 14., 16., 18.])
    >>> outfile = TemporaryFile()
    >>> x = mg.arange(10.0)
    >>> mg.save(outfile, x)
    >>> _ = outfile.seek(0) # Only needed here to simulate closing & reopening file
    >>> loaded = mg.load(outfile)
    >>> loaded
    Tensor([0., 1., 2., 3., 4., 5., 6., 7., 8., 9.])
    >>> loaded.grad
    array([ 0.,  2.,  4.,  6.,  8., 10., 12., 14., 16., 18.])
    """
    loaded = np.load(file)

    loaded_tensor = tb.tensor(loaded["data"])

    if "grad" in loaded:
        loaded_tensor.backward(loaded["grad"])

    return loaded_tensor



import numpy as np

from mygrad.tensor_base import asarray


def check_loss_inputs(x, y_true):
    """Ensures that the inputs to scores-truth style loss functions
    are of the correct shapes and types.

    Parameters
    ----------
    x : mygrad.Tensor, shape=(N, C)
        The C class scores for each of the N pieces of data.

    y_true : Sequence[int]
        The correct class-indices, in [0, C), for each datum.

    Raises
    ------
    TypeError
        `y_true` must be an integer-type array-like object
    ValueError
        `x` must be a 2-dimensional array-like object
        `y_true` must be a shape-(N,) array-like object
    """
    if not x.ndim == 2:
        raise ValueError(
            "`x` must be a 2-dimensional array-like object, got {}-dim".format(x.ndim)
        )

    y_true = asarray(y_true)
    if not np.issubdtype(y_true.dtype, np.integer):
        raise TypeError(
            "`y_true` must be an integer-type "
            "array-like object, got {}".format(y_true.dtype)
        )

    if y_true.ndim != 1 or y_true.shape[0] != x.shape[0]:
        raise ValueError(
            "`y_true` must be a shape-(N,) array: \n"
            "\tExpected shape-{}\n"
            "\tGot shape-{}".format((x.shape[0],), y_true.shape)
        )



import sys
from abc import ABC, ABCMeta, abstractmethod
from inspect import signature
from typing import (
    TYPE_CHECKING,
    Callable,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_type_hints,
    overload,
)

import numpy as np

from mygrad.operation_base import BinaryUfunc, Operation, Ufunc, UnaryUfunc, _NoValue
from mygrad.tensor_base import _REGISTERED_UFUNC, Tensor
from mygrad.typing import ArrayLike, DTypeLikeReals, Index, Mask, Real

__all__ = ["ufunc_creator"]


if sys.version_info >= (3, 8):  # pragma: no cover
    from typing import Literal

    HAS_LITERAL = True
else:  # pragma: no cover
    try:
        from typing_extensions import Literal
    except ImportError:
        HAS_LITERAL = False
    else:
        HAS_LITERAL = True

if TYPE_CHECKING and HAS_LITERAL:  # pragma: no cover
    One = Literal[1]
    Two = Literal[2]
else:  # pragma: no cover
    One = int
    Two = int


T = TypeVar("T")


def _permitted_type_str(x: str) -> bool:
    for code in ["F", "D", "G", "M", "m", "O"]:
        # filter non bool/int/float types
        if code in x:
            return False
    return True


class MyGradUfunc(type):
    _wrapped_op: Union[Type[UnaryUfunc], Type[BinaryUfunc]]
    _decorated_func: Callable
    nin: int
    nout: int
    nargs: int
    ntypes: int
    types: List[str]
    identity: int
    signature: str

    def __call__(cls, *args, **kwargs) -> Tensor:  # pragma: no cover
        raise NotImplementedError()

    def __repr__(cls) -> str:
        return f"<mygrad-ufunc '{cls._decorated_func.__name__}'>"


class Final(type):
    def __new__(mcs, name, bases, classdict):
        for b in bases:
            if isinstance(b, Final):
                raise TypeError("mygrad.ufunc cannot be subclassed")
        return type.__new__(mcs, name, bases, dict(classdict))


class FinalABC(ABCMeta, Final):
    pass


class ufunc(ABC, metaclass=FinalABC):
    """A un-inheritable, abstract superclass of all MyGrad ufuncs.

    Examples
    --------
    >>> import mygrad as mg
    >>> issubclass(mg.add, mg.ufunc)
    True
    """

    nin: int
    nout: int
    nargs: int
    ntypes: int
    types: List[str]
    identity: int
    signature: str

    def __call__(
        self,
        *args: ArrayLike,
        dtype: DTypeLikeReals,
        constant: Optional[bool] = None,
        **kwargs,
    ) -> Tensor:  # pragma: no cover
        pass

    def at(
        self,
        a: ArrayLike,
        indices: Union[ArrayLike, Index, Tuple[ArrayLike, Index]],
        b: Optional[ArrayLike] = None,
        *,
        constant: Optional[bool] = None,
    ) -> Tensor:  # pragma: no cover
        """Not implemented"""
        raise NotImplementedError()

    def accumulate(
        self,
        array: ArrayLike,
        axis: int = 0,
        dtype: DTypeLikeReals = None,
        out: Optional[Union[Tensor, np.ndarray]] = None,
        *,
        constant: Optional[bool] = None,
    ) -> Tensor:  # pragma: no cover
        """Not implemented"""
        raise NotImplementedError()

    def outer(
        self,
        a: ArrayLike,
        b: ArrayLike,
        *,
        dtype: DTypeLikeReals,
        out: Optional[Union[Tensor, np.ndarray]],
    ) -> Tensor:  # pragma: no cover
        """Not Implemented"""
        raise NotImplementedError()

    def reduce(
        self,
        a: ArrayLike,
        axis: Optional[Union[int, Tuple[int, ...]]] = 0,
        dtype: DTypeLikeReals = None,
        out: Optional[Union[Tensor, np.ndarray]] = None,
        keepdims: bool = False,
        initial: Real = _NoValue,
        where: Mask = True,
    ) -> Tensor:  # pragma: no cover
        """Not Implemented"""
        raise NotImplementedError()

    @abstractmethod
    def reduceat(
        self,
        a: ArrayLike,
        indices: ArrayLike,
        axis: Optional[Union[int, Tuple[int, ...]]] = 0,
        dtype: DTypeLikeReals = None,
        out: Optional[Union[Tensor, np.ndarray]] = None,
    ) -> Tensor:  # pragma: no cover
        """Not Implemented"""
        raise NotImplementedError()


class MyGradUnaryUfunc(MyGradUfunc):
    _wrapped_op: Type[UnaryUfunc]
    nin: One

    def __call__(
        cls,
        x: ArrayLike,
        out: Optional[Union[Tensor, np.ndarray]] = None,
        *,
        where: Mask = True,
        dtype: DTypeLikeReals = None,
        constant: Optional[bool] = None,
        **kwargs,
    ) -> Tensor:
        # it is fastest to check if out is None, which is likely the
        # most common scenario, and this is a very "hot path" in the
        # code
        if out is not None and isinstance(out, Tensor):
            out._in_place_op(
                cls._wrapped_op,
                x,
                op_kwargs={"where": where, "dtype": dtype, **kwargs},
                constant=constant,
            )
            return out
        else:
            return Tensor._op(
                cls._wrapped_op,
                x,
                op_kwargs={"where": where, "dtype": dtype, **kwargs},
                constant=constant,
                out=out,
            )


class MyGradBinaryUfunc(MyGradUfunc):
    _wrapped_op: Type[BinaryUfunc]
    nin: Two

    def __call__(
        cls,
        x: ArrayLike,
        y: ArrayLike,
        out: Optional[Union[Tensor, np.ndarray]] = None,
        *,
        where: Mask = True,
        dtype: DTypeLikeReals = None,
        constant: Optional[bool] = None,
    ) -> Tensor:
        # it is fastest to check if out is None, which is likely the
        # most common scenario, and this is a very "hot path" in the
        # code
        if out is not None and isinstance(out, Tensor):
            out._in_place_op(
                cls._wrapped_op,
                x,
                y,
                op_kwargs={"where": where, "dtype": dtype},
                constant=constant,
            )
            return out
        else:
            return Tensor._op(
                cls._wrapped_op,
                x,
                y,
                op_kwargs={"where": where, "dtype": dtype},
                constant=constant,
                out=out,
            )


class MyGradBinaryUfuncNoMask(MyGradUfunc):
    _wrapped_op: Type[BinaryUfunc]
    nin: Two

    def __call__(
        cls,
        x: ArrayLike,
        y: ArrayLike,
        out: Optional[Union[Tensor, np.ndarray]] = None,
        *,
        dtype: DTypeLikeReals = None,
        constant: Optional[bool] = None,
    ) -> Tensor:
        # it is fastest to check if out is None, which is likely the
        # most common scenario, and this is a very "hot path" in the
        # code
        if out is not None and isinstance(out, Tensor):
            out._in_place_op(
                cls._wrapped_op,
                x,
                y,
                op_kwargs={"dtype": dtype},
                constant=constant,
            )
            return out
        else:
            return Tensor._op(
                cls._wrapped_op,
                x,
                y,
                op_kwargs={"dtype": dtype},
                constant=constant,
                out=out,
            )


@overload
def _create_ufunc(
    op: Type[UnaryUfunc],
    decorated_func: Callable,
    *,
    at_op: Optional[Type[Operation]] = None,
    accumulate_op: Optional[Type[Operation]] = None,
    outer_op: Optional[Type[Operation]] = None,
    reduce_op: Optional[Type[Operation]] = None,
    reduceat_op: Optional[Type[Operation]] = None,
) -> MyGradUnaryUfunc:  # pragma: no cover
    ...


@overload
def _create_ufunc(
    op: Type[BinaryUfunc],
    decorated_func: Callable,
    *,
    at_op: Optional[Type[Operation]] = None,
    accumulate_op: Optional[Type[Operation]] = None,
    outer_op: Optional[Type[Operation]] = None,
    reduce_op: Optional[Type[Operation]] = None,
    reduceat_op: Optional[Type[Operation]] = None,
) -> MyGradBinaryUfunc:  # pragma: no cover
    ...


# noinspection PyPep8Naming
def _create_ufunc(
    op: Type[Ufunc],
    decorated_func,
    *,
    at_op=None,
    accumulate_op=None,
    outer_op=None,
    reduce_op=None,
    reduceat_op=None,
) -> Type[ufunc]:
    def at(
        a: ArrayLike,
        indices: Union[ArrayLike, Index, Tuple[ArrayLike, Index]],
        b: Optional[ArrayLike] = None,
        *,
        constant: Optional[bool] = None,
    ) -> Tensor:  # pragma: no cover
        """Not implemented"""
        raise NotImplementedError()

    def accumulate(
        array: ArrayLike,
        axis: int = 0,
        dtype: Optional[DTypeLikeReals] = None,
        out: Optional[Union[Tensor, np.ndarray]] = None,
        *,
        constant: Optional[bool] = None,
    ) -> Tensor:  # pragma: no cover
        """Not implemented"""
        raise NotImplementedError()

    def outer(
        a: ArrayLike,
        b: ArrayLike,
        *,
        dtype: Optional[DTypeLikeReals] = None,
        out: Optional[Union[Tensor, np.ndarray]] = None,
    ) -> Tensor:  # pragma: no cover
        """Not Implemented"""
        raise NotImplementedError()

    def reduce(
        a: ArrayLike,
        axis: Optional[Union[int, Tuple[int, ...]]] = 0,
        dtype: Optional[DTypeLikeReals] = None,
        out: Optional[Union[Tensor, np.ndarray]] = None,
        keepdims: bool = False,
        initial: Real = _NoValue,
        where: Mask = True,
    ) -> Tensor:  # pragma: no cover
        """Not Implemented"""
        raise NotImplementedError()

    def reduceat(
        a: ArrayLike,
        indices: ArrayLike,
        axis: Optional[Union[int, Tuple[int, ...]]] = 0,
        dtype: Optional[DTypeLikeReals] = None,
        out: Optional[Union[Tensor, np.ndarray]] = None,
    ) -> Tensor:  # pragma: no cover
        """Not Implemented"""
        raise NotImplementedError()

    if op.numpy_ufunc.nin == 1:
        MetaBuilder = MyGradUnaryUfunc
    elif op.numpy_ufunc.nin == 2:
        MetaBuilder = (
            MyGradBinaryUfunc if op._supports_where else MyGradBinaryUfuncNoMask
        )
    else:  # pragma: no cover
        raise NotImplementedError(
            "MyGrad Internal: `mygrad._utils.op_creator` only supports unary and binary ufuncs presently"
        )

    # filter out non-real dtypes
    types = [
        type_code
        for type_code in op.numpy_ufunc.types
        if _permitted_type_str(type_code)
    ]
    out = MetaBuilder(
        decorated_func.__name__,
        (object,),
        (
            {
                "_wrapped_op": op,
                "at": at,
                "accumulate": accumulate,
                "reduce": reduce,
                "reduceat": reduceat,
                "outer": outer,
                "signature": op.numpy_ufunc.signature,
                "identity": op.numpy_ufunc.identity,
                "nargs": op.numpy_ufunc.nargs,
                "nin": op.numpy_ufunc.nin,
                "nout": op.numpy_ufunc.nout,
                "ntypes": len(types),
                "types": types,
                "_decorated_func": decorated_func,
                "__name__": decorated_func.__name__,
                "__qualname__": decorated_func.__name__,
                "__signature__": signature(decorated_func),
                "__annotations__": get_type_hints(decorated_func),
                "__doc__": decorated_func.__doc__,
            }
        ),
    )
    ufunc.register(out)
    return out


class ufunc_creator:
    """Wraps function stub and returns a mygrad-ufunc with the same name and
    signature.

    Examples
    --------
    >>> from mygrad.ufuncs import ufunc_creator, MyGradUnaryUfunc
    >>> from mygrad.math.trigonometric.ops import Sin
    >>> @ufunc_creator(mygrad_op=Sin)
    ... def sin(x, *, constant=None):
    ...    '''custom docs'''

    ``sin`` is now a mygrad ufunc

    >>> isinstance(sin, MyGradUnaryUfunc)
    True
    >>> sin.nin  # `sin` has all of the attributes of a numpy-ufunc
    1
    >>>
    Tensor(0.)

    It has the docstring and signature that was specified by the above stub.

    >>> sin.__doc__
    custom docs
    >>> import inspect
    >>> inspect.signature(sin)
    <Signature (x, *, constant=None)>
    """

    def __init__(
        self,
        mygrad_op: Type[Ufunc],
        *,
        at_op: Optional[Type[Operation]] = None,
        accumulate_op: Optional[Type[Operation]] = None,
        outer_op: Optional[Type[Operation]] = None,
        reduce_op: Optional[Type[Operation]] = None,
        reduceat_op: Optional[Type[Operation]] = None,
    ):
        """Provide the Operation-based implementation for the ufunc, and, optionally, implementations
        for the various methods of that ufunc (e.g. ufunc.at, ufunc.outer, etc.)

        Parameters
        ----------
        mygrad_op : Type[Ufunc]
            An operation-based implementation of a ufunc that supports back-propagation
            through mygrad tensors.
        """
        if not issubclass(mygrad_op, (UnaryUfunc, BinaryUfunc)):
            raise TypeError(
                "ufunc_creator can only accept `UnaryUfunc` and `BinaryUfunc` operations"
            )
        self.op = mygrad_op
        if any(
            item is not None
            for item in (at_op, accumulate_op, outer_op, reduce_op, reduceat_op)
        ):
            raise NotImplementedError(
                "There isn't support for binding ufunc methods presently."
            )

        self.at_op = at_op
        self.accumulate_op = accumulate_op
        self.outer_op = outer_op
        self.reduce_op = reduce_op
        self.reduceat_op = reduceat_op

    def __call__(self, decorated_func: T) -> T:
        out_ufunc = _create_ufunc(
            self.op,
            decorated_func=decorated_func,
            at_op=self.at_op,
            accumulate_op=self.accumulate_op,
            outer_op=self.outer_op,
            reduce_op=self.reduce_op,
            reduceat_op=self.reduceat_op,
        )
        _REGISTERED_UFUNC[getattr(np, out_ufunc.__name__)] = out_ufunc
        return out_ufunc



from numbers import Real
from typing import Optional

from numpy import ndarray

from mygrad.math.misc.funcs import maximum, minimum
from mygrad.tensor_base import Tensor
from mygrad.typing import ArrayLike

__all__ = ["hard_tanh"]


def hard_tanh(
    x: ArrayLike,
    *,
    lower_bound: Real = -1,
    upper_bound: Real = 1,
    constant: Optional[bool] = None,
) -> Tensor:
    """Returns the hard hyperbolic tangent function.

    The hard_tanh function is `lower_bound` where `x` <= `lower_bound`, `upper_bound` where
    `x` >= `upper_bound`, and `x` where `lower_bound` < `x` < `upper_bound`.

    Parameters
    ----------
    x : ArrayLike
        The input, to which to apply the hard tanh function.

    lower_bound : Real, optional (default=-1)
        The lower bound on the hard tanh.

    upper_bound : Real, optional (default=1)
        The upper bound on the hard tanh.

    constant : Optional[bool]
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient).

    Returns
    -------
    mygrad.Tensor
        The result of applying the "hard-tanh" function elementwise to `x`.

    Examples
    --------
    >>> import mygrad as mg
    >>> from mygrad.nnet.activations import hard_tanh
    >>> x = mg.arange(-5, 6)
    >>> x
    Tensor([-5, -4, -3, -2, -1,  0,  1,  2,  3,  4,  5])
    >>> y = hard_tanh(x, lower_bound=-3, upper_bound=3); y
    Tensor([-3, -3, -3, -2, -1,  0,  1,  2,  3,  3,  3])
    >>> y.backward()
    >>> x.grad
    array([0., 0., 0., 1., 1., 1., 1., 1., 0., 0., 0.])

    .. plot::

       >>> import mygrad as mg
       >>> from mygrad.nnet.activations import hard_tanh
       >>> import matplotlib.pyplot as plt
       >>> x = mg.linspace(-6, 6, 100)
       >>> y = hard_tanh(x, lower_bound=-3, upper_bound=3)
       >>> plt.title("hard_tanh(x, lower_bound=-3, upper_bound=3)")
       >>> y.backward()
       >>> plt.plot(x, x.grad, label="df/dx")
       >>> plt.plot(x, y, label="f(x)")
       >>> plt.legend()
       >>> plt.grid()
       >>> plt.show()
    """
    if isinstance(lower_bound, (ndarray, Tensor)):
        lower_bound = lower_bound.item()

    if isinstance(upper_bound, (ndarray, Tensor)):
        upper_bound = upper_bound.item()

    if not isinstance(lower_bound, Real):
        raise TypeError(
            f"`lower_bound` must be a real-valued scalar, got {lower_bound} (type { type(lower_bound)})"
        )

    if not isinstance(upper_bound, Real):
        raise TypeError(
            f"`upper_bound` must be a real-valued scalar, got {upper_bound} (type {type(upper_bound)})"
        )

    return maximum(
        lower_bound, minimum(x, upper_bound, constant=constant), constant=constant
    )



import numpy as np

from mygrad.operation_base import Operation

__all__ = ["Tensor_Transpose_Property", "Transpose", "MoveAxis", "Roll", "SwapAxes"]


class Tensor_Transpose_Property(Operation):
    can_return_view = True

    def __call__(self, a):
        """Same as a.transpose(), except that a is returned if
        a.ndim < 2.

        Parameters
        ----------
        a : mygrad.Tensor"""
        self.variables = (a,)
        return a.data.T

    def backward_var(self, grad, index, **kwargs):
        return grad.T


class Transpose(Operation):
    can_return_view = True

    def __call__(self, a, axes=None):
        self.variables = (a,)
        if axes is not None:
            self.axes = tuple(axis % a.ndim for axis in axes)
        else:
            self.axes = tuple(range(a.ndim)[::-1])
        return np.transpose(a.data, axes)

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        if a.ndim > 1:
            grad = grad.transpose(np.argsort(self.axes))
        return grad


class MoveAxis(Operation):
    can_return_view = True

    def __call__(self, a, source, destination):
        self.variables = (a,)
        self.source = source
        self.destination = destination
        return np.moveaxis(a.data, source, destination)

    def backward_var(self, grad, index, **kwargs):
        if not index == 0:  # pragma: no cover
            raise IndexError
        return np.moveaxis(grad, self.destination, self.source)


class SwapAxes(Operation):
    can_return_view = True

    def __call__(self, a, axis1, axis2):
        self.variables = (a,)
        self.axis1 = axis1
        self.axis2 = axis2
        return np.swapaxes(a.data, axis1, axis2)

    def backward_var(self, grad, index, **kwargs):
        if not index == 0:  # pragma: no cover
            raise IndexError
        return np.swapaxes(grad, self.axis2, self.axis1)


class Roll(Operation):
    def __call__(self, a, shift, axis):
        self.variables = (a,)
        self.shift = shift
        self.axis = axis
        return np.roll(a.data, shift=shift, axis=axis)

    def backward_var(self, grad, index, **kwargs):
        if not index == 0:  # pragma: no cover
            raise IndexError
        rev_shift = (
            -self.shift
            if not hasattr(self.shift, "__iter__")
            else tuple(-i for i in self.shift)
        )
        return np.roll(grad, axis=self.axis, shift=rev_shift)



from typing import Optional

import numpy as np

import mygrad._utils.graph_tracking as _tracking
from mygrad.math._special import logsumexp
from mygrad.operation_base import Operation
from mygrad.tensor_base import Tensor
from mygrad.typing import ArrayLike

from ._utils import check_loss_inputs


class SoftmaxCrossEntropy(Operation):
    """Given the classification scores of C classes for N pieces of data,
    computes the NxC softmax classification probabilities. The
    cross entropy is then computed by using the true classification labels.

    log-softmax is used for improved numerical stability"""

    def __call__(self, x, y_true):
        """Parameters
        ----------
        x : mygrad.Tensor, shape=(N, C)
            The C class scores for each of the N pieces of data.

        y_true : Sequence[int]
            The correct class-indices, in [0, C), for each datum.

        Returns
        -------
        The average softmax loss"""
        if isinstance(y_true, Tensor):
            y_true = y_true.data

        check_loss_inputs(x, y_true)
        self.variables = (x,)
        scores = x.data
        log_softmax = scores - logsumexp(scores, axis=-1, keepdims=True)
        label_locs = (range(len(scores)), y_true)
        loss = -np.sum(log_softmax[label_locs]) / scores.shape[0]

        if _tracking.TRACK_GRAPH:
            self.back = np.exp(log_softmax)
            self.back[label_locs] -= 1.0
            self.back /= scores.shape[0]
        return loss

    def backward_var(self, grad, index, **kwargs):
        return grad * self.back


def softmax_crossentropy(
    x: ArrayLike, y_true: ArrayLike, *, constant: Optional[bool] = None
) -> Tensor:
    r"""Given the classification scores of C classes for N pieces of data,

    computes the NxC softmax classification probabilities. The
    cross entropy is then computed by using the true classification labels.

    log-softmax is used for improved numerical stability.

    Parameters
    ----------
    x : ArrayLike, shape=(N, C)
        The C class scores for each of the N pieces of data.

    y_true : ArrayLike, shape=(N,)
        The correct class-indices, in [0, C), for each datum.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    The average softmax loss

    Raises
    ------
    ValueError
        Bad dimensionalities for ``x`` or ``y_true``

    Notes
    -----
    - :math:`N` is the number of samples in the batch.
    - :math:`C` is the number of possible classes for which scores are provided.

    Given the shape-:math:`(N, C)` tensor of scores, ``x``, the softmax classification
    probabilities are computed. That is, the score for class-:math:`k` of a given datum
    (:math:`s_{k}`) is normalized using the 'softmax' transformation:

    .. math::

        p_{k} = \frac{e^{s_k}}{\sum_{i=1}^{C}{e^{s_i}}}

    This produces the "prediction probability distribution", :math:`p`, for each datum.
    The cross-entropy loss for that datum is then computed according to the true class-index
    for that datum, as reported in ``y_true``. That is the "true probability distribution",
    :math:`t`, for the datum is :math:`1` for the correct class-index and :math:`0` elsewhere.

    The cross-entropy loss for that datum is thus:

    .. math::
       l = - \sum_{k=1}^{C}{t_{k} \log{p_{k}}}

    Having computed each per-datum cross entropy loss, this function then returns the loss
    averaged over all :math:`N` pieces of data:

    .. math::

       L = \frac{1}{N}\sum_{i=1}^{N}{l_{i}}

    Examples
    --------
    >>> import mygrad as mg
    >>> from mygrad.nnet import softmax_crossentropy

    Let's take a simple case where N=1, and C=3. We'll thus make up classification
    scores for a single datum. Suppose the scores are identical for the three classes
    and that the true class is class-0:

    >>> x = mg.Tensor([[2., 2., 2.]])  # a shape-(1, 3) tensor of scores
    >>> y_true = mg.Tensor([0])  # the correct class for this datum is class-0

    Because the scores are identical for all three classes, the softmax normalization
    will simply produce :math:`p = [\frac{1}{3}, \frac{1}{3}, \frac{1}{3}]`. Because
    class-0 is the "true" class, :math:`t = [1., 0., 0.]`. Thus our softmax cross-entropy
    loss should be:

    .. math::
      -(1 \times \log{\frac{1}{3}} + 0 \times \log{\frac{1}{3}} + 0 \times \log{\frac{1}{3}})
      = \log(3) \approx 1.099

    Let's see that this is what ``softmax_crossentropy`` returns:

    >>> softmax_crossentropy(x, y_true)
    Tensor(1.09861229)

    Similarly, suppose a datum's scores are :math:`[0, 0, 10^6]`, then the softmax normalization
    will return :math:`p \approx [0., 0., 1.]`. If the true class for this datum is class-2, then
    the loss should be nearly 0, since :math:`p` and :math:`t` are essentially identical:

    .. math::
      -(0 \times \log{0} + 0 \times \log{0} + 1 \times \log{1})
      = -\log(1) = 0

    Now, let's construct ``x`` and ``y_true`` so that they incorporate the scores/labels for
    both of the data that we have considered:

    >>> x = mg.Tensor([[2., 2.,  2.],  # a shape-(2, 3) tensor of scores
    ...                [0., 0., 1E6]])
    >>> y_true = mg.Tensor([0, 2])     # the class IDs for the two data

    ``softmax_crossentropy(x, y_true)`` will return the average loss of these two data,
    :math:`\frac{1}{2}(1.099 + 0) \approx 0.55`:

    >>> softmax_crossentropy(x, y_true)
    Tensor(0.54930614)
    """
    return Tensor._op(SoftmaxCrossEntropy, x, op_args=(y_true,), constant=constant)



"""
Exposes MyGrad's ufunc types (which are distinct from the Operation subclasses that are provided to
implement the machinery for backpropagation through a particular operation). Also provides a wrapper
for creating a ufunc with a specified name and signature from a given function stub."""

from ._ufunc_creators import MyGradBinaryUfunc, MyGradUnaryUfunc, ufunc_creator

__all__ = ["ufunc_creator", "MyGradUnaryUfunc", "MyGradBinaryUfunc"]



from mygrad.tensor_creation.funcs import identity

from .constant import constant
from .dirac import dirac
from .glorot_normal import glorot_normal
from .glorot_uniform import glorot_uniform
from .he_normal import he_normal
from .he_uniform import he_uniform
from .normal import normal
from .uniform import uniform


__all__ = [
    "constant",
    "dirac",
    "glorot_normal",
    "glorot_uniform",
    "he_normal",
    "he_uniform",
    "identity",
    "normal",
    "uniform",
]



"""
Defines the base class for mathematical operations capable of back-propagating
gradients to their input tensors."""
from abc import ABC, abstractmethod
from numbers import Real
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Set, Tuple, Union
from weakref import ReferenceType

import numpy as np

from mygrad._utils import SkipGradient, reduce_broadcast
from mygrad.errors import InvalidBackprop, InvalidGradient
from mygrad.typing import DTypeLike, Mask

if TYPE_CHECKING:  # pragma: no cover
    from mygrad import Tensor
    from mygrad._utils import WeakRef

__all__ = [
    "Operation",
    "Ufunc",
    "UnaryUfunc",
    "BinaryUfunc",
    "Sequential",
]

Axis = Optional[Union[int, Tuple[int, ...]]]


class _NoValueType:
    """Special keyword value.

    The instance of this class may be used as the default value assigned to a
    deprecated keyword in order to check if it has been given a user defined
    value.
    """

    __instance = None

    def __new__(cls):
        # ensure that only one instance exists
        if not cls.__instance:
            cls.__instance = super(_NoValueType, cls).__new__(cls)
        return cls.__instance

    def __repr__(self):  # pragma: no cover
        return "<no value>"


_NoValue = _NoValueType()


class Operation(ABC):
    """Base class for all tensor operations that support back-propagation
    of gradients.

    Consider the Operation-instance ``f``. A forward-pass through ``f`` is defined
    via ``f.__call__(...)``. Thus, given tensors ``a`` and ``b``, a computational
    graph is defined ``f.__call__(a, b) -> c``, where the "creator" of tensor ``c``
    is recorded as ``f``::

          (node: a) --+
                       -> [operation: f(a, b)] --> (node: c)
          (node: b) --+

    Back-propagating through ``c`` will instruct ``f`` to back-propagate
    the gradient to its inputs, which are recorded as ``a`` and ``b``. Each
    node then back-propagates to any Operation-instance that is recorded
    as its creator, and so on.
    """

    # Can be set to true if the operation is guaranteed to not returns a view
    # this will reduce some overhead on checking for shared memory
    can_return_view = False  # type: bool

    # Stores the input tensors that the operation will backprop through.
    variables: Tuple["Tensor", ...]

    def __init__(self):
        # Stores positional and keyword arguments used to call op.
        # Can be set optionally - only if op needs to be "replayed",
        # e.g. with a view
        self.replay_args: Optional[Tuple[Any, ...]] = None
        self.replay_kwargs: Optional[Dict[str, Any]] = None
        self.replay_force_constant: Optional[bool] = None
        self.where: Mask = True

    @staticmethod
    def grad_post_process_fn(
        grad: np.ndarray, var_shape: Tuple[int, ...]
    ) -> np.ndarray:
        # this function gets called all of the time; we can avoid
        # the extra function call by doing the shape check upfront
        if grad.shape == var_shape:
            return grad
        out = reduce_broadcast(grad, var_shape)

        if out.ndim == 0:
            # sum-reduction to a scalar produces a float
            out = np.array(out, copy=False)
        return out

    @abstractmethod
    def __call__(self, *input_vars: "Tensor", **kwargs) -> np.ndarray:
        """Performs a forward pass, f, of this Operation::

                         f(x1, ...., xn)

        Parameters
        ----------
        *input_vars : mygrad.Tensor
            The input-arguments of f. The tuple (x1, ...., xn)
            should be bound to the instance-attribute `self.variables`

        **kwargs : Any
            Additional arguments for the operation

        Returns
        -------
        numpy.ndarray
            The output of the forward pass function.

        Notes
        -----
        This method should set the ``self.variables`` attribute
        with a tuple storing all of the input tensors of this operations"""

        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def backward_var(self, grad: np.ndarray, index: int, **kwargs) -> np.ndarray:
        """Given ``grad = dℒ/df``, computes ``∂ℒ/∂x_{i}``, where ``x_{i}`` is one
        of ``x1, ...., xn``.

        ``ℒ`` is assumed to be the terminal node from which ``ℒ.backward()`` was
        called.

        Parameters
        ----------
        grad : numpy.ndarray
            The back-propagated total derivative with respect to the present
            operation: dℒ/df. This will have the same shape as f, the result
            of the forward pass.

        index : int
            The index-location of ``var`` in ``self.variables``

        Returns
        -------
        numpy.ndarray
            ∂ℒ/∂x_{i}

        Raises
        ------
        SkipGradient"""
        raise NotImplementedError()  # pragma: no cover

    def backward(
        self,
        grad: np.ndarray,
        *,
        graph: Set["WeakRef[Operation]"],
        **kwargs,
    ):
        """Back-propagates the gradient through all of the operation's inputs,
        which are stored in the tuple `self.variables`.

        Constant tensors (`tensor.constant is True`) skipped by this process.

        Parameters
        ----------
        grad : numpy.ndarray
            The back-propagated total derivative with respect to the present
            operation (`f`): d(out)/df

        graph : Set[Operation]
            The set of all operations relevant to the terminal node of the computational graph,
            which triggered back-propagation.
        """
        for index, var in enumerate(self.variables):
            if not var.constant:
                if not var._ops:
                    raise InvalidBackprop(
                        f"Part of the computational graph containing "
                        f"this tensor, {var}, was 'cleared' prior to backprop.\n"
                        f"It is recommended that you clear all computational graphs "
                        f"and restart your computation."
                    )

                try:
                    # don't cast to array here so that we have an easier time
                    # doing type checking (e.g. avoid `None` -> `array(None, dtype=obj)`
                    backed_grad = self.backward_var(grad, index, **kwargs)
                except SkipGradient:
                    continue

                if not isinstance(backed_grad, (np.ndarray, np.number, Real)):
                    raise InvalidGradient(
                        f"An invalid gradient-value was passed to:"
                        f"\n\t`{type(self).__name__}.backward_var(<gradient>, index={index})`"
                        f"\nGradients are expected to be real-valued scalars or "
                        f"numpy arrays, got a gradient of type: {type(backed_grad)}"
                    )

                backed_grad = np.array(backed_grad, copy=False)

                if self.where is not True:
                    backed_grad = backed_grad * self.where

                backed_grad = self.grad_post_process_fn(backed_grad, var.shape)
                assert backed_grad.shape == var.shape, (backed_grad.shape, var.shape)
                if var._grad is None:
                    backed_grad = (
                        np.copy(backed_grad)
                        # `backed_grad` is view of grad; we want to be able to
                        # augment tmp-grad inplace later
                        if backed_grad.base is not None or (backed_grad is grad)
                        else backed_grad
                    )
                    if backed_grad.dtype != var.dtype:
                        backed_grad = backed_grad.astype(var.dtype, copy=False)

                    var._grad = backed_grad
                else:
                    var._grad += backed_grad

        # Avoid visiting the same node multiple times. Note that we don't store
        # these by the node itself, since Tensors are unhashable, but by its `id`.
        visited = set()
        ref_op = ReferenceType(self)

        for var in (
            i for i in self.variables if not i.constant and i.creator is not None
        ):
            var_id = id(var)
            if var_id in visited:
                continue
            visited.add(var_id)
            var._accum_ops.add(ref_op)
            var._backward(graph=graph)


class Ufunc(Operation, ABC):
    """The base class for mygrad's universal functions.

    'A universal function (or ufunc for short) is a function that operates on
    ndarrays in an element-by-element fashion, supporting array broadcasting, type casting,
    and several other standard features. That is, a ufunc is a “vectorized” wrapper for a
    function that takes a fixed number of specific inputs and produces a fixed number of
    specific outputs.' [1]_

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/ufuncs.html"""

    numpy_ufunc: np.ufunc
    _supports_where: bool = True


class UnaryUfunc(Ufunc, ABC):
    """A base class that specifies the common interface to – and facilitates
    back-prop through – ufuncs that operate on a single array argument;
    e.g. `mygrad.sin`, `mygrad.negative`."""

    def __call__(
        self,
        x1: "Tensor",
        out: Optional[np.ndarray] = None,
        *,
        where: Mask = True,
        dtype: DTypeLike = None,
    ) -> np.ndarray:
        """f(x1, out=None, *, where=True, dtype=None)

        Parameters
        ----------
        x1 : Tensor, shape-(...)
            The input to the operation.

            This tensor is saved to the state of the operation instance
            so that back-prop can be performed through it.

        out : Optional[np.ndarray]
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or None,
            a freshly-allocated array is returned.

        where: Union[bool, np.ndarray]
            Accepts a boolean array which is broadcast together with ``x1``.
            Values of True indicate to calculate the ufunc at that position, values
            of False indicate to leave the value in the output alone.

        dtype : Optional[numpy.dtype, str, object]
            Overrides the dtype of the calculation and output array.

        Returns
        -------
        y : ndarray, shape-(...)
            A numpy array of the same shape as ``x1`` with the ufunc applied
            elementwise on ``x1``.

        Notes
        -----
        This docstring was adapted from numpy's documentation [1]_.

        References
        ----------
        .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.sqrt.html
        """
        self.variables: Tuple["Tensor"] = (x1,)
        if where is not True:
            self.where = where
        return self.numpy_ufunc(x1.data, out=out, where=where, dtype=dtype)


class BinaryUfunc(Ufunc, ABC):
    """A base class that specifies the common interface to – and facilitates
    back-prop through – mygrad's ufuncs that operate on a two array arguments;
    e.g. `mygrad.add`, `mygrad.multiply`.
    """

    def __call__(
        self,
        x1: "Tensor",
        x2: "Tensor",
        out: Optional[np.ndarray] = None,
        *,
        where: Mask = True,
        dtype: DTypeLike = None,
    ) -> np.ndarray:
        """f(x1, x2, out=None, *, where=True, dtype=None)

        Parameters
        ----------
        x1 : Tensor
            The first input to the operation.

            This tensor is saved to the state of the operation instance
            so that back-prop can be performed through it.

        x2 : Tensor
            The second input to the operation.

            This tensor is saved to the state of the operation instance
            so that back-prop can be performed through it.

        out : Optional[np.ndarray]
            A location into which the result is stored. If provided, it must
            have a shape that the inputs broadcast to. If not provided or None,
            a freshly-allocated array is returned.

        where: Union[bool, np.ndarray]
            Accepts a boolean array which is broadcast jointly with ``x1`` and ``x2``.
            Values of True indicate to calculate the ufunc at that position, values
            of False indicate to leave the value in the output alone.

        dtype : Optional[numpy.dtype, str, object]
            Overrides the dtype of the calculation and output array.

        Returns
        -------
        y : ndarray
            A numpy array resulting from the elementwise application of the ufunc to
            corresponding pairs of elements from ``x1`` and ``x2``, respectively.

            If ``x1`` and ``x2`` are of different shapes, then the operation is broadcast
            across them [1]_.

        Notes
        -----
        This docstring was adapted from numpy's documentation [2]_.

        References
        ----------
        .. [1] https://numpy.org/doc/stable/user/basics.broadcasting.html
        .. [2] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.add.html
        """
        self.variables: Tuple["Tensor", "Tensor"] = (x1, x2)
        if where is not True and where is not _NoValue:
            self.where = where
            return self.numpy_ufunc(x1.data, x2.data, out=out, where=where, dtype=dtype)
        else:
            return self.numpy_ufunc(x1.data, x2.data, out=out, dtype=dtype)


class Sequential(Operation, ABC):
    """A base class that specifies the common interface to – and facilitates
    back-prop through – numpy's sequential functions; e.g. `numpy.sum`, `numpy.var`,
    `numpy.max`"""

    numpy_func: Callable[..., np.ndarray]
    _integer_axis_only: bool = False

    @staticmethod
    @abstractmethod
    def numpy_func(
        a: np.ndarray,
        axis: Optional[Union[int, Tuple[int, ...]]] = None,
        dtype: DTypeLike = None,
        out: Optional[np.ndarray] = None,
        *args,
        **kwargs,
    ) -> np.ndarray:
        raise NotImplementedError()  # pragma: no cover

    def __init__(self):
        self.axis: Axis
        self.keepdims: Optional[bool]
        self.initial: Real
        self.out_shape: Tuple[int, ...]
        super().__init__()

    def __call__(
        self,
        a: "Tensor",
        axis: Axis = None,
        dtype=None,
        out: Optional[np.ndarray] = None,
        keepdims: bool = _NoValue,
        initial: Real = _NoValue,
        *,
        where: Union[bool, np.ndarray] = _NoValue,
        ddof: int = _NoValue,
    ) -> np.ndarray:
        self.variables: Tuple["Tensor"] = (a,)

        if where is not True and where is not _NoValue:
            self.where = where

        self.keepdims = keepdims
        self.initial = initial
        self.ddof = ddof

        # Unless axis is None or the op is integer-axis-only
        # normalize axis to be a tuple of ints.
        if (
            not self._integer_axis_only
            and axis is not None
            and not hasattr(axis, "__iter__")
        ):
            self.axis = (axis,)
        else:
            self.axis = axis

        kwargs = {}

        if keepdims is not _NoValue:
            kwargs["keepdims"] = keepdims

        if initial is not _NoValue:  # pragma: no cover
            kwargs["initial"] = initial

        if where is not _NoValue:
            kwargs["where"] = where

        if ddof is not _NoValue:
            kwargs["ddof"] = ddof

        if dtype is not _NoValue:
            kwargs["dtype"] = dtype

        out = self.numpy_func(a.data, axis=axis, out=out, **kwargs)
        self.out_shape = out.shape

        return out



from typing import Optional

import numpy as np

from mygrad.operation_base import Operation
from mygrad.tensor_base import Tensor
from mygrad.typing import ArrayLike


class ReLu(Operation):
    def __call__(self, a):
        self.variables = (a,)
        self.back = np.asarray(a > 0, dtype=a.dtype)
        return a.data * self.back

    def backward_var(self, grad, index, **kwargs):
        return grad * self.back


def relu(x: ArrayLike, *, constant: Optional[bool] = None) -> Tensor:
    """
    Applies the recitfied linear unit activation function::

        f(x) = {x, x > 0
                0, x <= 0 }

    Parameters
    ----------
    x : ArrayLike
        relu is applied element-wise on ``x``.

    constant : Optional[bool]
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    mygrad.Tensor

    Examples
    --------
    >>> import mygrad as mg
    >>> from mygrad.nnet import relu
    >>> x = mg.linspace(-5, 5, 5)
    >>> x
    Tensor([-5. , -2.5,  0. ,  2.5,  5. ])
    >>> relu(x)
    Tensor([-0. , -0. ,  0. ,  2.5,  5. ])
    >>> relu(x).backward()
    >>> x.grad  # d(relu(x))/dx
    array([0., 0., 0., 1., 1.])

    .. plot::

       >>> import mygrad as mg
       >>> from mygrad.nnet.activations import relu
       >>> import matplotlib.pyplot as plt
       >>> x = mg.linspace(-2, 2, 100)
       >>> y = relu(x)
       >>> plt.title("relu(x)")
       >>> y.backward()
       >>> plt.plot(x, x.grad, label="df/dx")
       >>> plt.plot(x, y, label="f(x)")
       >>> plt.legend()
       >>> plt.grid()
       >>> plt.show()
    """
    return Tensor._op(ReLu, x, constant=constant)



from numbers import Number

import numpy as np

import mygrad._utils.graph_tracking as _tracking
from mygrad.operation_base import Operation

__all__ = ["GetItem", "SetItem"]


def _is_int_array_index(index):
    """Returns True if `index` contains any array-like integer-valued sequences

    Parameters
    ----------
    index : Tuple[Any]

    Returns
    -------
    bool"""
    return any(
        np.issubdtype(np.asarray(ind).dtype, np.int_) and np.asarray(ind).ndim
        for ind in index
    )


def _is_bool_array_index(index):
    """Returns True if `index` solely contains a boolean-valued array

    Parameters
    ----------
    index : Tuple[Any]

    Returns
    -------
    bool"""
    return len(index) == 1 and np.issubdtype(np.asarray(index[0]).dtype, np.bool_)


class GetItem(Operation):
    """Defines the __getitem__ interface for a Tensor, supporting back-propagation

    Supports back-propagation through all valid numpy-indexing (basic, advanced, mixed, etc.)"""

    can_return_view = True

    def __call__(self, a, index):
        """``a[index]``

        Parameters
        ----------
        a : mygrad.Tensor
            The tensor whose entries are being accessed.

        index : valid-array-index
            An n-dimensional index for specifying entries or subregions of `a`.
            All means of numpy-array indexing (basic, advanced, mixed, etc) are
            supported.

        Returns
        -------
        numpy.ndarray
            The array returned by the get-item operation"""
        self.variables = (a,)
        self.index = index if isinstance(index, tuple) else (index,)
        out = a.data[index]

        self._used_distinct_indices = (
            (
                out.base is not None
                and (out.base is a.data or out.base is a.data.base)
                or out.ndim == 0
                or isinstance(out, Number)
                or _is_bool_array_index(self.index)
            )
            if _tracking.TRACK_GRAPH
            else None
        )
        return out

    def backward_var(self, grad, index, **kwargs):
        a = self.variables[index]
        out = np.zeros_like(a.data)  # TODO: ensure ints are promoted to floats
        if self._used_distinct_indices:
            out[self.index] += grad
        else:
            # although `add.at` will work for all cases, it is
            # a very slow function: https://github.com/numpy/numpy/issues/5922
            np.add.at(out, self.index, grad)
        return out


def _arr(*shape: int) -> np.ndarray:
    """Construct an array of a specified consisting of values [0, _arr.size)
    filled in row-major order.

    Parameters
    ----------
    *shape : int

    Returns
    -------
    numpy.ndarray"""
    return np.arange(np.prod(shape)).reshape(shape)


class SetItem(Operation):
    """Defines the __setitem__ interface for a Tensor, supporting back-propagation through
    both the tensor being set and the tensor whose .

    Supports back-propagation through all valid numpy-indexing (basic, advanced, mixed, etc.),
    as well as"""

    can_return_view = True

    def __call__(self, a, b, index, *, out: np.ndarray):
        """a[index] = b

        Parameters
        ----------
        a : mygrad.Tensor
            The tensor whose entries are being set. A copy of the underlying
            data is made if `a` is a non-constant tensor.

        b : mygrad.Tensor
            `b` must be broadcast-compatible with `a[index]`

        index : valid-array-index
            An n-dimensional index for specifying entries or subregions of `a`.
            All means of numpy-array indexing (basic, advanced, mixed, etc) are
            supported.

        out : ndarray
            The data to be mutated; defaults to ``a.data``. This argument is
            included to provide parity with standard ufunc signatures.

        Notes
        -----
        Additional computational overhead is required for back-propagation when
        `index` contains any integer-valued arrays, to accommodate for the scenario
        in which a single element is set multiple times."""

        self.variables = (a, b)
        self.index = index if isinstance(index, tuple) else (index,)
        out[index] = b.data
        return out

    def backward_var(self, grad, index, **kwargs):
        a, b = self.variables
        if index == 0:
            grad = np.copy(grad)
            grad[self.index] = 0
            return grad
        elif index == 1:
            grad_sel = np.asarray(grad[self.index])

            # Basic indexing and indexing with a single boolean-array is trivial. The
            # gradient into b can just be accessed by indexing into `grad`.
            # Indexing with integer-valued arrays can be problematic, as the same
            # item can be specified multiple for "setting"; here only the last set-item
            # for that element has an effect. For example:
            #     x[np.array([0, 0])] = np.array([2, 3])  # `3` gets set to x[0]; 2 has no effect
            # Thus only that corresponding element in `grad` (that corresponding to `3`)
            # should be propagated back into b. Thus we must check to see if any items are
            # being set redundantly, and mask out any elements in `grad` corresponding to
            # the elements in `b` that weren't actually set.
            if (
                (grad_sel.base is None or not np.shares_memory(grad_sel, grad))
                and grad_sel.size > 0
                and grad_sel.ndim > 0
                and not _is_bool_array_index(self.index)
                and _is_int_array_index(self.index)
            ):
                # create an array of unique elements, and see if indexing into it produces
                # any redundant elements
                unique = _arr(*grad.shape)
                sub_sel = unique[self.index].flat
                (
                    elements,
                    first_inds,
                ) = np.unique(np.flip(sub_sel, axis=0), return_index=True)
                if len(first_inds) < len(sub_sel):
                    # one or more elements were set redundantly, identify the entries in `b`
                    # that actually were set to those elements (the last-most set-item calls
                    # for those elements) and propagate only the corresponding elements from grad

                    first_inds = (len(sub_sel) - 1) - first_inds
                    mask = np.zeros_like(sub_sel)
                    mask[first_inds] = 1
                    mask = mask.reshape(grad_sel.shape)
                    grad_sel *= mask

            # handle the edge case of "projecting down" on setitem. E.g:
            # x = Tensor([0, 1, 2])
            # y = Tensor([3])
            # x[0] = y  # this is legal since x[0] and y have the same size
            if grad_sel.ndim < b.ndim:
                if grad_sel.size == b.size:
                    grad_sel = grad_sel.reshape(b.shape)
                else:
                    # Broadcasting occurred during set-item and `b` contains
                    # excess leading singleton dimensions. Make `grad_sel`
                    # commensurate with `b` for subsequent `reduce_broadcast`
                    # to work
                    grad_sel = grad_sel[(np.newaxis,) * (b.ndim - grad_sel.ndim)]

            return grad_sel
        else:
            raise IndexError()  # pragma: no cover
